; file: "server.scm", 11/10/96

; Test use of C-interface to write a "Scheme server", that is a program
; which is mainly written in C and that makes calls to Scheme functions to
; access certain services.

;------------------------------------------------------------------------------

; C part.

; This part is typically in a separate C source file which is compiled
; independently from the server and combined with the server at
; the final link.  Here we put the C code in a "c-declare" form
; so that all the code of the program is in this file.

(c-declare "

extern char *eval_string (); /* function to access in Scheme library */

#include <stdio.h>
#include <stdlib.h>

extern ___mod_or_lnk ____20_program__ (); /* mangled name of Scheme library */

int main (argc, argv)
int argc;
char **argv;
{
  char *temp;

  /* setup the Scheme library */

  ___setup_params setup_params;

  setup_params.argc                = argc;
  setup_params.argv                = argv;
  setup_params.min_heap            = 0;
  setup_params.max_heap            = 0;
  setup_params.gc_hook             = 0;
  setup_params.fatal_error         = 0;
  setup_params.standard            = 0;
  setup_params.debug_level         = 0;
  setup_params.default_io_encoding = ___IO_CHAR_ENCODING; /* native encoding */
  setup_params.force_tty           = 0;
  setup_params.linker              = ____20_program__;

  ___setup (&setup_params);

  /* main part of program: call Scheme functions */

  temp = eval_string (\"(define x 200)\");
  if (temp != 0)
    {
      printf (\"result = %s\\n\", temp);
      ___EXT (___free_string (temp));
    }

  temp = eval_string (\"(expt 2 x)\");
  if (temp != 0)
    {
      printf (\"result = %s\\n\", temp);
      ___EXT (___free_string (temp));
    }

  temp = eval_string (\"(+ 1 2\"); /* note: missing closing parenthesis */
  if (temp != 0)
    {
      printf (\"result = %s\\n\", temp);
      ___EXT (___free_string (temp));
    }

  temp = eval_string (\"(+ x y)\"); /* note: y is unbound */
  if (temp != 0)
    {
      printf (\"result = %s\\n\", temp);
      ___EXT (___free_string (temp));
    }

  /* cleanup the Scheme library (required before terminating) */

  ___cleanup ();

  /* terminate program */

  exit (0);
}
")

;------------------------------------------------------------------------------

; Scheme part.

(define (catch-all-errors thunk)
  (call-with-current-continuation
    (lambda (return)
      (##catch-all
        (lambda (signal-name args)
          (return
            (if (symbol? signal-name) (symbol->string signal-name) #f)))
        thunk))))

(define (read-from-string str)
  (with-input-from-string str read))

(define (write-to-string obj)
  (with-output-to-string (lambda () (write obj))))

(c-define (eval-string str) (char-string) char-string "eval_string" "extern"
  (catch-all-errors
    (lambda () (write-to-string (eval (read-from-string str))))))

;------------------------------------------------------------------------------
