;==============================================================================

; file: "_t-c-3.scm"

; Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved.

(include "generic.scm")

;------------------------------------------------------------------------------
;
; Back end for C language (part 3)
; -----------------------

(define targ-bits-per-byte 8) ; 8 bits per byte

; FIXNUM representation.

; Limits of fixnums (assumes 2's complement representation):

(define targ-min-fixnum
  (* -2
     (expt 2 (- (* targ-min-word-size targ-bits-per-byte)
                (+ targ-tag-bits 2)))))

(define targ-max-fixnum
  (- (+ targ-min-fixnum 1)))

(define (targ-fixnum? x)
  (and (>= x targ-min-fixnum)
       (<= x targ-max-fixnum)))

; BIGNUM representation.

; Radix must be a power of two and targ-bignum-radix^2 - 1 must fit in
; a fixnum.

(define targ-bignum-radix
  (expt 2
        (quotient (- (* targ-min-word-size targ-bits-per-byte)
                     (+ targ-tag-bits 1))
                  2)))

(define (targ-bignum-digits obj)

  (define (integer->digits n)
    (if (= n 0)
      '()
      (cons (remainder n targ-bignum-radix)
            (integer->digits (quotient n targ-bignum-radix)))))

  (if (< obj 0)
    (cons 0 (integer->digits (- obj)))
    (cons 1 (integer->digits obj))))

; FLONUM representation.

(define targ-inexact-+2   (exact->inexact 2))
(define targ-inexact--2   (exact->inexact -2))
(define targ-inexact-+1   (exact->inexact 1))
(define targ-inexact-+1/2 (exact->inexact (/ 1 2)))
(define targ-inexact-0    (exact->inexact 0))

(define (targ-float->inexact-exponential-format x f64?)
  (let* ((e-bits (if f64? 11 8))
         (e-bias (- (expt 2 (- e-bits 1)) 1)))

    (define (exp-form-pos x y i)
      (let ((i*2 (+ i i)))
        (let ((z (if (and (not (< e-bias i*2))
                          (not (< x y)))
                   (exp-form-pos x (* y y) i*2)
                   (cons x 0))))
          (let ((a (car z)) (b (cdr z)))
            (let ((i+b (+ i b)))
              (if (and (not (< e-bias i+b))
                       (not (< a y)))
                (begin
                  (set-car! z (/ a y))
                  (set-cdr! z i+b)))
              z)))))

    (define (exp-form-neg x y i)
      (let ((i*2 (+ i i)))
        (let ((z (if (and (< i*2 (- e-bias 1))
                          (< x y))
                   (exp-form-neg x (* y y) i*2)
                   (cons x 0))))
          (let ((a (car z)) (b (cdr z)))
            (let ((i+b (+ i b)))
              (if (and (< i+b (- e-bias 1))
                       (< a y))
                (begin
                  (set-car! z (/ a y))
                  (set-cdr! z i+b)))
              z)))))

    (define (exp-form x)
      (if (< x targ-inexact-+1)
        (let ((z (exp-form-neg x targ-inexact-+1/2 1)))
          (set-car! z (* targ-inexact-+2 (car z)))
          (set-cdr! z (- -1 (cdr z)))
          z)
        (exp-form-pos x targ-inexact-+2 1)))

    (if (negative? x)
      (let ((z (exp-form (- targ-inexact-0 x))))
        (set-car! z (- targ-inexact-0 (car z)))
        z)
      (exp-form x))))

(define (targ-float->exact-exponential-format x f64?)
  (let* ((z      (targ-float->inexact-exponential-format x f64?))
         (m-bits (if f64? 52 23))
         (e-bits (if f64? 11 8)))

    (let ((y (car z)))
      (cond ((not (< y targ-inexact-+2))
             (set-car! z (expt 2 m-bits))
             (set-cdr! z (expt 2 (- e-bits 1))))
            ((not (< targ-inexact--2 y))
             (set-car! z (- 0 (expt 2 m-bits)))
             (set-cdr! z (expt 2 (- e-bits 1))))
            (else
             (set-car! z
               (truncate
                 (inexact->exact
                   (* (car z) (exact->inexact (expt 2 m-bits))))))))
      (set-cdr! z (- (cdr z) m-bits))
      z)))

(define (targ-float->bits x f64?)
  (let ((m-bits (if f64? 52 23))
        (e-bits (if f64? 11 8)))

    (define (bits a b)
      (let ((m-min (expt 2 m-bits)))
        (if (< a m-min)
          a
          (+ (- a m-min)
             (* (+ (+ b m-bits) (- (expt 2 (- e-bits 1)) 1))
                m-min)))))

    (let ((z (targ-float->exact-exponential-format x f64?)))
      (let ((a (car z)) (b (cdr z)))
        (if (negative? a)
          (+ (expt 2 (+ e-bits m-bits)) (bits (- 0 a) b))
          (bits a b))))))

(define (targ-f32->bits x)
  (targ-float->bits x #f))

(define (targ-f64->hi-lo-bits x)
  (let ((bits (targ-float->bits x #t)))
    (cons (quotient bits #x100000000)
          (modulo   bits #x100000000))))

; RATNUM representation.

(define (targ-numerator x)
  (numerator x))

(define (targ-denominator x)
  (denominator x))

; CPXNUM representation.

(define (targ-real-part x)
  (real-part x))

(define (targ-imag-part x)
  (imag-part x))

; Extraction of object's type and subtype.

(define (targ-obj-type obj)
  (cond ((false-object? obj)
         'boolean)
        ((eq? obj #t)
         'boolean)
        ((null? obj)
         'null)
        ((absent-object? obj)
         'absent)
        ((void-object? obj)
         'void)
        ((end-of-file-object? obj)
         'eof)
        ((optional-object? obj)
         'optional)
        ((rest-object? obj)
         'rest)
        ((key-object? obj)
         'key)
        ((script-object? obj)
         'script)
        ((symbol-object? obj)
         'subtyped)
        ((keyword-object? obj)
         'subtyped)
        ((proc-obj? obj)
         'procedure)
        ((pair? obj)
         'pair)
        ((number? obj)
         (cond ((and (integer? obj) (exact? obj) (targ-fixnum? obj))
                'fixnum)
               (else
                'subtyped)))
        ((char? obj)
         'char)
        (else
         'subtyped)))

(define (targ-obj-subtype obj)
  (cond ((symbol-object? obj)
         'symbol)
        ((keyword-object? obj)
         'keyword)
        ((number? obj)
         (cond ((and (integer? obj) (exact? obj))
                'bignum)
               ((and (rational? obj) (exact? obj))
                'ratnum)
               ((and (inexact? (targ-real-part obj))
                     (exact? (targ-imag-part obj))
                     (zero? (targ-imag-part obj)))
                'flonum)
               (else
                'cpxnum)))
        ((u8vect? obj)
         'u8vector)
        ((u16vect? obj)
         'u16vector)
        ((u32vect? obj)
         'u32vector)
        ((f32vect? obj)
         'f32vector)
        ((f64vect? obj)
         'f64vector)
        ((vector? obj)
         'vector)
        ((string? obj)
         'string)
        (else
         (compiler-internal-error
           "targ-obj-subtype, unknown object 'obj'" obj))))
