;==============================================================================

; file: "gsc.scm"

; Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved.

(##include "generic.scm")

;------------------------------------------------------------------------------

(set! make-global-environment ; import runtime macros into compilation env
  (lambda ()

    (define (extract-macros cte)
      (if (##cte-top? cte)
        (env-frame #f '())
        (let ((parent-cte (##cte-parent-cte cte)))
          (if (##cte-macro? cte)
            (env-macro (extract-macros parent-cte)
                       (##cte-macro-name cte)
                       (##cte-macro-def cte))
            (extract-macros parent-cte)))))

    (extract-macros (##cte-top-cte ##interaction-cte))))

(define (compile-file-to-c
         filename
         #!optional (options (absent-obj)) (output (absent-obj)))
  (force-vars (filename)
    (check-string filename (compile-file-to-c filename options output)
      (let ((opts
             (if (##eq? options (absent-obj))
               '()
               (force-vars (options)
                 options))))
        (if (##eq? output (absent-obj))
          (c#cf filename #f opts #f)
          (force-vars (output)
            (check-string output (compile-file-to-c filename options output)
              (c#cf filename #f opts output))))))))

(define (compile-file
         filename
         #!optional (options (absent-obj)))
  (force-vars (filename)
    (check-string filename (compile-file filename options)
      (let ((dynamic-cc ##dynamic-cc))
        (if (##procedure? dynamic-cc)

          (let* ((opts
                  (if (##eq? options (absent-obj))
                    '()
                    (force-vars (options)
                      options)))
                 (x
                  (c#cf filename #f opts #f)))
            (if x
              (let ((root (##path-strip-extension filename)))
                (let loop ((version 1))
                  (let* ((root-obj
                          (##string-append
                            root
                            ".o"
                            (##number->string version 10)))
                         (root-obj-port
                          (##open-input-file root-obj)))
                    (if root-obj-port
                      (begin
                        (##close-port root-obj-port)
                        (loop (##fixnum.+ version 1)))
                      (let ((msg (dynamic-cc root root-obj)))
                        (if msg
                          (##runtime-error
                            msg
                            'compile-file
                            (##cons filename options))
                          (##void)))))))
              (##void)))

          (##runtime-error
            "Dynamic loading is not available on this platform"
            'compile-file
            (##cons filename options)))))))

(define (link-incremental
          modules
          #!optional (output (absent-obj)) (base (absent-obj)))
  (if (##not (##pair? modules))
    (check-string modules (link-incremental modules output base)
      #f)
    (let loop ((lst modules) (rev-mods '()))
      (force-vars (lst)
        (if (##pair? lst)
          (let ((s (##car lst)))
            (force-vars (s)
              (check-string s (link-incremental modules output base)
                (loop (##cdr lst) (##cons s rev-mods)))))
          (let* ((baselib
                  (if (##eq? base (absent-obj))
                    (##string-append (or (##path-expand "~~") "") "_gambc")
                    (force-vars (base)
                      base)))
                 (baselib-and-mods
                  (##cons baselib (##reverse rev-mods))))
            (check-string baselib (link-incremental modules output base)
              (if (##eq? output (absent-obj))
                (c#targ-linker #t
                               baselib-and-mods
                               #f
                               (##string-append (##car rev-mods) "_"))
                (force-vars (output)
                  (check-string output (link-incremental modules output base)
                    (c#targ-linker #t
                                   baselib-and-mods
                                   output
                                   #f)))))))))))

(define (link-flat modules #!optional (output (absent-obj)))
  (if (##not (##pair? modules))
    (check-string modules (link-flat modules output)
      #f)
    (let loop ((lst modules) (rev-mods '()))
      (force-vars (lst)
        (if (##pair? lst)
          (let ((s (##car lst)))
            (force-vars (s)
              (check-string s (link-flat modules output)
                (loop (##cdr lst) (##cons s rev-mods)))))
          (let ((mods (##reverse rev-mods)))
            (if (##eq? output (absent-obj))
              (c#targ-linker #f
                             mods
                             #f
                             (##string-append (##car rev-mods) "_"))
              (force-vars (output)
                (check-string output (link-flat modules output)
                  (c#targ-linker #f
                                 mods
                                 output
                                 #f))))))))))

;------------------------------------------------------------------------------

(define (##main)

  (##define-macro (gsi-or x) x)

  (define (process-initialization-file)

    (define (try filename)
      (let ((init (##open-input-file filename)))
        (if init
          (begin
            (##close-port init)
            (##eval-top (##list '##include filename) ##interaction-cte)
            #t)
          #f)))

    (or (try "gambc.scm")
        (let ((homedir (##path-expand "~")))
          (if homedir
            (try (##string-append homedir "gambc.scm"))))))

  (define (interpreter-batch-mode arguments)
    (let ((loaded-script? #f))
      (let loop ((lst arguments))
        (if (##not (or loaded-script? (##null? lst)))
          (begin
            (##load (##car lst)
                    (lambda (script?)
                      (if script?
                        (begin
                          (set! loaded-script? #t)
                          (set! ##processed-argv lst))))
                    #t
                    #f)
            (loop (##cdr lst)))
          (##exit 0)))))

  (define (compiler-batch-mode options arguments)

    (define (scm-file? file)
      (##member (##path-extension file) '(".scm" "")))

    (define (c-file? file)
      (##equal? (##path-extension file) ".c" #f))

    (let* ((opts (##map ##car options))
           (sym-opts (##map ##string->symbol opts)))

      (let loop1 ((lst arguments)
                  (nb-scm-files 0))
        (if (##not (##null? lst))

          (let ((arg (##car lst)))
            (loop1 (##cdr lst)
                   (if (scm-file? arg)
                     (##fixnum.+ nb-scm-files 1)
                     nb-scm-files)))

          (let* ((no-link?
                  (##member "c" opts))
                 (output
                  (let ((x (##assoc "o" options)))
                    (cond ((##not x)
                           #f)
                          ((and no-link? (##fixnum.< 1 nb-scm-files))
                           (##write-string
                            "*** WARNING -- Multiple output files: \"o\" option ignored"
                            ##stderr)
                           (##newline ##stderr)
                           #f)
                          (else
                           (##cdr x))))))

            (let loop2 ((lst arguments)
                        (rev-roots '()))
              (if (##not (##null? lst))

                (let* ((file (##car lst))
                       (root (##path-strip-extension file)))
                  (cond ((scm-file? file)
                         (if (##fixnum.< 1 nb-scm-files)
                           (begin
                             (##display file ##stdout #f)
                             (##display ":" ##stdout #f)
                             (##newline ##stdout)))
                         (if (##not (c#cf file #f sym-opts
                                          (and no-link? output)))
                           (##exit 1))
                       (loop2 (##cdr lst) (##cons root rev-roots)))
                      ((c-file? file)
                       (loop2 (##cdr lst) (##cons root rev-roots)))
                      (else
                       (##display file ##stdout #f)
                       (##display ": ignored (file not recognized)" ##stdout #f)
                       (##newline ##stdout)
                       (loop2 (##cdr lst) rev-roots))))

              (let* ((flat?
                      (##member "flat" opts))
                     (base
                      (let ((x (##assoc "l" options)))
                        (cond ((##not x)
                               #f)
                              ((or no-link? flat?)
                               (##write-string
                                "*** WARNING -- No incremental link: \"l\" option ignored"
                                ##stderr)
                               (##newline ##stderr)
                               #f)
                              (else
                               (##cdr x))))))

                (if (##member "c" opts)
                  (if flat?
                    (begin
                      (##write-string
                        "*** WARNING -- \"c\" option was specified: \"flat\" option ignored"
                        ##stderr)
                      (##newline ##stderr)))
                  (if (##not (##null? rev-roots))
                    (let ((roots (##reverse rev-roots)))
                      (if flat?
                        (link-flat
                          roots
                          (or output (absent-obj)))
                        (link-incremental
                          roots
                          (or output (absent-obj))
                          (or base (absent-obj)))))))))))))))

  (define (interactive-mode)

    (##define-macro (banner)
      (string-append "Gambit Version " c#compiler-version))

    (##display (banner) ##stdout #f)
    (##newline ##stdout)
    (##newline ##stdout)
    (##repl))

  (define (pipe-mode)
    (let ((port ##stdin))
      (let loop ()
        (let ((expr (##read port)))
          (if (##eof-object? expr)
            (##exit 0)
            (let ((val (##eval-top expr ##interaction-cte)))
              (##write val ##stdout #t)
              (##newline ##stdout)
              (loop)))))))

  (define (##split-command-line options-with-arguments cont)

    (define (option? arg)
      (and (##fixnum.< 1 (##string-length arg))
           (##char=? (##string-ref arg 0) #\-)))

    (define (convert-option arg)
      (##substring arg 1 (##string-length arg)))

    (let loop1 ((arguments (##cdr (##argv)))
                (rev-options '()))
      (if (and (##pair? arguments) (option? (##car arguments)))

        (let ((opt (convert-option (##car arguments)))
              (rest (##cdr arguments)))
          (if (##member opt options-with-arguments)
            (if (##pair? rest)
              (loop1 (##cdr rest)
                     (##cons (##cons opt (##car rest)) rev-options))
              (begin
                (##write-string
                  "*** WARNING -- Missing argument for option \""
                  ##stderr)
                (##display opt ##stderr #f)
                (##write-string "\"" ##stderr)
                (##newline ##stderr)
                (loop1 rest rev-options)))
            (loop1 rest
                   (##cons (##cons opt #f) rev-options))))

        (cont (##reverse rev-options) arguments))))

  (define (split-options lst options)
    (if (##null? lst)
      '(())
      (let* ((x (##car lst))
             (opt (##car x)))
        (if (##member opt options)
          (let ((y (split-options (##cdr lst) options)))
            (##cons (##cons x (##car y)) (##cdr y)))
          (##cons '() lst)))))

  (let ((gambcdir (##path-expand "~~")))
    (if gambcdir
      (##load (##string-append gambcdir "gambc") (lambda (script?) #f) #f #f)))

  (##split-command-line
    '("o" "l")
    (lambda (raw-options arguments)
      (let* ((options1 (split-options raw-options '("f" "i")))
             (main-options (##car options1))
             (other-options (##cdr options1))
             (skip-initialization-file? (##assoc "f" main-options))
             (force-interpreter? (##assoc "i" main-options))
             (unprocessed-options
              (if (gsi-or force-interpreter?)
                other-options
                (##cdr (split-options
                        other-options
                        '("c" "flat" "o" "l"
                          "verbose" "report" "expansion" "gvm"
                          "check" "force" "debug"))))))
        (if (##not (##null? unprocessed-options))
          (begin
            (##write-string "*** WARNING -- Unprocessed options: " ##stderr)
            (##write (##map ##car unprocessed-options) ##stderr #f)
            (##newline ##stderr)))
        (if (##not skip-initialization-file?)
          (process-initialization-file))
        (cond ((##not (##null? arguments))
               (if (gsi-or force-interpreter?)
                 (interpreter-batch-mode arguments)
                 (compiler-batch-mode other-options arguments)))
              ((##port-isatty ##stdin)
               (interactive-mode))
              (else
               (pipe-mode)))))))

(##main)
