;==============================================================================

; file: "_io.scm"

; Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

(##declare (not interrupts-enabled))

;------------------------------------------------------------------------------

; Generic port objects.

(##define-macro
  (make-port

    input?        ; #t if is an input port or an input-output port
    output?       ; #t if is an output port or an input-output port

    name          ; name associated with port (file name, window name, etc)

    read-char     ; (lambda (port) ...)
    peek-char     ; (lambda (port) ...)
    char-ready?   ; (lambda (port) ...)
    flush-input   ; (lambda (port) ...)
    read-pos      ; (lambda (port) ...)

    write-char    ; (lambda (char port) ...)
    flush-output  ; (lambda (port) ...)
    write-pos     ; (lambda (port) ...)

    close         ; (lambda (port) ...)

    width         ; (lambda (port) ...)

    isatty        ; (lambda (port) ...)

    . others      ; other slots
  )
  `(##subtype-set!
     (##vector
       #t ; port is open
       ,input?
       ,output?
       ,name
       ,read-char
       ,peek-char
       ,char-ready?
       ,flush-input
       ,read-pos
       ,write-char
       ,flush-output
       ,write-pos
       ,close
       ,width
       ,isatty
       ,@others)
     (subtype-port)))

(##define-macro (port-open? port)
  `(##vector-ref ,port 0))

(##define-macro (port-open?-set! port x)
  `(##vector-set! ,port 0 ,x))

(##define-macro (port-input? port)
  `(##vector-ref ,port 1))

(##define-macro (port-output? port)
  `(##vector-ref ,port 2))

(##define-macro (port-name port)
  `(##vector-ref ,port 3))

(##define-macro (port-read-char port)
  `((##vector-ref ,port 4) ,port))

(##define-macro (port-peek-char port)
  `((##vector-ref ,port 5) ,port))

(##define-macro (port-char-ready? port)
  `((##vector-ref ,port 6) ,port))

(##define-macro (port-flush-input port)
  `((##vector-ref ,port 7) ,port))

(##define-macro (port-read-pos port)
  `((##vector-ref ,port 8) ,port))

(##define-macro (port-write-char char port)
  `((##vector-ref ,port 9) ,char ,port))

(##define-macro (port-flush-output port)
  `((##vector-ref ,port 10) ,port))

(##define-macro (port-write-pos port)
  `((##vector-ref ,port 11) ,port))

(##define-macro (port-close port)
  `((##vector-ref ,port 12) ,port))

(##define-macro (port-width port)
  `((##vector-ref ,port 13) ,port))

(##define-macro (port-isatty port)
  `((##vector-ref ,port 14) ,port))

(##define-macro (port-input-char-count port)
  `(##vector-ref ,port 15))

(##define-macro (port-input-char-count-set! port pos)
  `(##vector-set! ,port 15 ,pos))

(##define-macro (port-input-line-count port)
  `(##vector-ref ,port 16))

(##define-macro (port-input-line-count-set! port pos)
  `(##vector-set! ,port 16 ,pos))

(##define-macro (port-input-line-start port)
  `(##vector-ref ,port 17))

(##define-macro (port-input-line-start-set! port pos)
  `(##vector-set! ,port 17 ,pos))

(##define-macro (port-output-char-count port)
  `(##vector-ref ,port 18))

(##define-macro (port-output-char-count-set! port pos)
  `(##vector-set! ,port 18 ,pos))

(##define-macro (port-output-line-count port)
  `(##vector-ref ,port 19))

(##define-macro (port-output-line-count-set! port pos)
  `(##vector-set! ,port 19 ,pos))

(##define-macro (port-output-line-start port)
  `(##vector-ref ,port 20))

(##define-macro (port-output-line-start-set! port pos)
  `(##vector-set! ,port 20 ,pos))

(define (##open-port? port)
  (port-open? port))

(define (##input-port? port)
  (port-input? port))

(define (##output-port? port)
  (port-output? port))

;------------------------------------------------------------------------------

; Position tracking for ports.

(##define-macro (exact-int.add1 n) ; Add 1 to a fixnum or bignum.
  `(let ((temp ,n))
     (if (and (##fixnum? temp) (##fixnum.< temp (max-fixnum)))
       (##fixnum.+ temp 1)
       (##+ temp 1))))

(##define-macro (end-of-line? char)
  `(##char=? ,char #\newline))

(##define-macro (track-input-pos port char)
  `(if (##char? ,char)
     (let ((char-count (exact-int.add1 (port-input-char-count ,port))))
       (port-input-char-count-set! ,port char-count)
       (if (end-of-line? ,char)
         (begin
           (port-input-line-start-set! ,port char-count)
           (port-input-line-count-set! ,port 
             (exact-int.add1 (port-input-line-count ,port))))))))

(##define-macro (track-output-pos port char)
  `(let ((char-count (exact-int.add1 (port-output-char-count ,port))))
     (port-output-char-count-set! ,port char-count)
     (if (end-of-line? ,char)
       (begin
         (port-output-line-start-set! ,port char-count)
         (port-output-line-count-set! ,port 
           (exact-int.add1 (port-output-line-count ,port)))))))

;------------------------------------------------------------------------------

; "Stdio" ports.

(c-declare "#include \"os.h\"")

(##define-macro (io-getc stream)
  `((c-lambda ((pointer "___STREAM")) scheme-object "___io_getc") ,stream))

(##define-macro (io-peek stream)
  `((c-lambda ((pointer "___STREAM")) scheme-object "___io_peek") ,stream))

(##define-macro (io-ready stream)
  `((c-lambda ((pointer "___STREAM")) scheme-object "___io_ready") ,stream))

(##define-macro (io-putc stream char)
  `((c-lambda ((pointer "___STREAM") ucs4) scheme-object "___io_putc") ,stream ,char))

(##define-macro (io-flush stream)
  `((c-lambda ((pointer "___STREAM")) scheme-object "___io_flush") ,stream))

(##define-macro (io-close stream)
  `((c-lambda ((pointer "___STREAM")) scheme-object "___io_close") ,stream))

(##define-macro (io-free stream)
  `((c-lambda ((pointer "___STREAM")) scheme-object "___io_free") ,stream))

(##define-macro (io-width stream)
  `((c-lambda ((pointer "___STREAM")) scheme-object "___io_width") ,stream))

(##define-macro (io-isatty stream)
  `((c-lambda ((pointer "___STREAM")) scheme-object "___io_isatty") ,stream))

(##define-macro (io-stdin)
  `((c-lambda () (pointer "___STREAM") "___io_stdin")))

(##define-macro (io-stdout)
  `((c-lambda () (pointer "___STREAM") "___io_stdout")))

(##define-macro (io-stderr)
  `((c-lambda () (pointer "___STREAM") "___io_stderr")))

(##define-macro (io-interrupted-call? x)
  `((c-lambda (scheme-object) scheme-object "___result = ___BOOL(___ARG1 == ___FIX((EINTR)));") ,x))

(##define-macro (io-open-input path encoding)
  `((c-lambda (char-string int int) (pointer "___STREAM") "___io_open")
    ,path
    -1
    ,encoding))

(##define-macro (io-open-output path encoding)
  `((c-lambda (char-string int int) (pointer "___STREAM") "___io_open")
    ,path
    1
    ,encoding))

(##define-macro (io-open-input-output path encoding)
  `((c-lambda (char-string int int) (pointer "___STREAM") "___io_open")
    ,path
    0
    ,encoding))

(define (##io-encoding char-encoding)
  (case char-encoding
    ((char)   1)
    ((latin1) 2)
    ((utf8)   3)
    ((byte)   4)
    ((ucs2)   5)
    ((ucs4)   6)
    (else     0)))

(define (##open-input-file-no-path-expand
          path
          #!optional (char-encoding #f))
  (let ((stream
         (io-open-input
          path
          (##io-encoding char-encoding))))
    (if stream
      (##make-io-port #t #f path stream stream)
      #f)))

(define (##open-input-file
          path
          #!optional (char-encoding #f))
  (let ((expanded-path (##path-expand path)))
    (and expanded-path
         (##open-input-file-no-path-expand expanded-path char-encoding))))

(define (##open-output-file-no-path-expand
          path
          #!optional (char-encoding #f))
  (let ((stream
         (io-open-output
          path
          (##io-encoding char-encoding))))
    (if stream
      (##make-io-port #f #t path stream stream)
      #f)))

(define (##open-output-file
          path
          #!optional (char-encoding #f))
  (let ((expanded-path (##path-expand path)))
    (and expanded-path
         (##open-output-file-no-path-expand expanded-path char-encoding))))

(define (##open-input-output-file-no-path-expand
          path
          #!optional (char-encoding #f))
  (let ((stream
         (io-open-input-output
          path
          (##io-encoding char-encoding))))
    (if stream
      (##make-io-port #t #t path stream stream)
      #f)))

(define (##open-input-output-file
          path
          #!optional (char-encoding #f))
  (let ((expanded-path (##path-expand path)))
    (and expanded-path
         (##open-input-output-file-no-path-expand expanded-path char-encoding))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##io-exception-handler port x continue)
  (cond ((io-interrupted-call? x)
         (let ()
           (##declare (interrupts-enabled))
           (continue)))
        (else
         (##signal '##signal.io-error "IO error on" port)
         (continue))))

(##define-macro (port-input-stream port)
  `(##vector-ref ,port 21))

(##define-macro (port-input-stream-set! port stream)
  `(##vector-set! ,port 21 ,stream))

(##define-macro (port-output-stream port)
  `(##vector-ref ,port 22))

(##define-macro (port-output-stream-set! port stream)
  `(##vector-set! ,port 22 ,stream))

(define (##make-io-port
          input?
          output?
          name
          input-stream
          output-stream)

  (define (stream-flush stream port)
    (let loop ()
      (let ((x (io-flush stream)))
        (if (##fixnum? x)
          (##io-exception-handler port x (lambda () (loop)))
          #f))))

  (define (stream-close stream port)
    (let loop ()
      (let ((x (io-close stream)))
        (if (##fixnum? x)
          (##io-exception-handler port x (lambda () (loop)))
          #f))))

  (make-port

    input?
    output?

    name

    ; read-char
    (lambda (port)
      (let ((stream (port-input-stream port)))
        (let loop ()
          (let ((c (io-getc stream)))
            (if (##fixnum? c)
              (##io-exception-handler port c (lambda () (loop)))
              (begin
                (track-input-pos port c)
                c))))))

    ; peek-char
    (lambda (port)
      (let ((stream (port-input-stream port)))
        (let loop ()
          (let ((c (io-peek stream)))
            (if (##fixnum? c)
              (##io-exception-handler port c (lambda () (loop)))
              c)))))

    ; char-ready?
    (lambda (port)
      (let ((stream (port-input-stream port)))
        (io-ready stream)))

    ; flush-input
    (lambda (port)
      (stream-flush (port-input-stream port) port))

    ; read-pos
    (lambda (port)
      (##vector (port-input-char-count port)
                (port-input-line-count port)
                (port-input-line-start port)))

    ; write-char
    (lambda (char port)
      (let ((stream (port-output-stream port)))
        (let loop ()
          (let ((x (io-putc stream char)))
            (if (##fixnum? x)
              (##io-exception-handler port x (lambda () (loop)))
              (begin
                (track-output-pos port char)
                #f))))))

    ; flush-output
    (lambda (port)
      (stream-flush (port-output-stream port) port))

    ; write-pos
    (lambda (port)
      (##vector (port-output-char-count port)
                (port-output-line-count port)
                (port-output-line-start port)))

    ; close
    (lambda (port)
      (let ((input-stream (port-input-stream port))
            (output-stream (port-output-stream port)))
        (stream-close input-stream port)
        (io-free input-stream)
        (if (##not (##eq? input-stream output-stream))
          (begin
            (stream-close output-stream port)
            (io-free output-stream)))
        (port-input-stream-set! port #f)
        (port-output-stream-set! port #f)
        (port-open?-set! port #f)
        (##void)))

    ; width
    (lambda (port)
      (let ((stream (port-output-stream port)))
        (io-width stream)))

    ; isatty
    (lambda (port)
      (let ((stream (port-input-stream port)))
        (io-isatty stream)))

    0 ; input-char-count
    0 ; input-line-count
    0 ; input-line-start
    0 ; output-char-count
    0 ; output-line-count
    0 ; output-line-start

    input-stream
    output-stream
    ))

(define ##stdin
  (let ((stream (io-stdin)))
    (##make-io-port #t #f 'stdin stream stream)))

(define ##stdout
  (let ((stream (io-stdout)))
    (##make-io-port #f #t 'stdout stream stream)))

(define ##stderr
  (let ((stream (io-stderr)))
    (##make-io-port #f #t 'stderr stream stream)))

;------------------------------------------------------------------------------

; String ports.

(##define-macro (port-string port)
  `(##vector-ref ,port 21))

(##define-macro (port-string-set! port str)
  `(##vector-set! ,port 21 ,str))

(define (##make-string-port input? output? str)

  ; assumes str is at most (max-fixnum) characters long

  (define (string-extend str)
    (let* ((n (##string-length str))
           (m (##fixnum.+ (##fixnum.* n 2) 1))
           (result (##make-string m #\space)))
      (let loop ((i (##fixnum.- n 1)))
        (if (##fixnum.< i 0)
          result
          (begin
            (##string-set! result i (##string-ref str i))
            (let ()
              (##declare (interrupts-enabled))
              (loop (##fixnum.- i 1))))))))

  (make-port

    input?
    output?

    'string

    ; read-char
    (lambda (port)
      (let ((n (port-output-char-count port))
            (i (port-input-char-count port)))
        (if (##fixnum.< i n)
          (let ((c (##string-ref (port-string port) i)))
            (track-input-pos port c)
            c)
          #!eof)))

    ; peek-char
    (lambda (port)
      (let ((n (port-output-char-count port))
            (i (port-input-char-count port)))
        (if (##fixnum.< i n)
          (##string-ref (port-string port) i)
          #!eof)))

    ; char-ready?
    (lambda (port)
      (let ((n (port-output-char-count port))
            (i (port-input-char-count port)))
        (##fixnum.< i n)))

    ; flush-input
    (lambda (port)
      #f)

    ; read-pos
    (lambda (port)
      (##vector (port-input-char-count port)
                (port-input-line-count port)
                (port-input-line-start port)))

    ; write-char
    (lambda (char port)
      (let* ((n (port-output-char-count port))
             (str (port-string port))
             (len (##string-length str)))
        (let ((s (if (##fixnum.< n len)
                   str
                   (let ((extended-str (string-extend str)))
                     (port-string-set! port extended-str)
                     extended-str))))
          (##string-set! s n char)
          (track-output-pos port char)
          #f)))

    ; flush-output
    (lambda (port)
      #f)

    ; write-pos
    (lambda (port)
      (##vector (port-output-char-count port)
                (port-output-line-count port)
                (port-output-line-start port)))

    ; close
    (lambda (port)
      (let* ((n (port-output-char-count port))
             (i (port-input-char-count port))
             (str (port-string port)))
        (port-open?-set! port #f)
        (if (##fixnum.= i 0)
          (begin
            (##string-shrink! str n)
            str)
          (##substring str i n))))

    ; width
    (lambda (port)
      79)

    ; isatty
    (lambda (port)
      #f)

    0 ; input-char-count
    0 ; input-line-count
    0 ; input-line-start
    (if str (##string-length str) 0) ; output-char-count
    0 ; output-line-count
    0 ; output-line-start

    (if str str (##make-string 127 #\space))
    ))

(define (##open-input-string str)
  (##make-string-port #t #f str))

(define (##open-output-string)
  (##make-string-port #f #t #f))

;------------------------------------------------------------------------------

; Generic port operations.

(define (##port-name port)
  (port-name port))

(define (##read-char port)
  (port-read-char port))

(define (##peek-char port)
  (port-peek-char port))

(define (##char-ready? port)
  (port-char-ready? port))

(define (##write-char c port)
  (port-write-char c port))

(define (##write-substring s i j port)
  (let loop ((i i))
    (if (##fixnum.< i j)
      (begin
        (port-write-char (##string-ref s i) port)
        (let ()
          (##declare (interrupts-enabled))
          (loop (##fixnum.+ i 1)))))))

(define (##write-string s port)
  (##write-substring s 0 (##string-length s) port))

(define (##close-port port)
  (port-close port))

(define (##newline port)
  (##write-char #\newline port))

(define (##port-width port)
  (port-width port))

(define (##port-isatty port)
  (port-isatty port))

(define (##flush-output port)
  (port-flush-output port))

;------------------------------------------------------------------------------

(##declare (interrupts-enabled))

(define (##read port)

  (##define-macro (+ . args)                `(##fixnum.+ ,@args))
  (##define-macro (- . args)                `(##fixnum.- ,@args))
  (##define-macro (= . args)                `(##fixnum.= ,@args))
  (##define-macro (< . args)                `(##fixnum.< ,@args))
  (##define-macro (exact? . args)           `(##exact? ,@args))
  (##define-macro (integer? . args)         `(##integer? ,@args))
  (##define-macro (real? . args)            `(##real? ,@args))
  (##define-macro (eq? . args)              `(##eq? ,@args))
  (##define-macro (assq . args)             `(##assq ,@args))
  (##define-macro (cdr . args)              `(##cdr ,@args))
  (##define-macro (unicode->character . args)`(##fixnum.->char ,@args))
  (##define-macro (char-alphabetic? . args) `(##char-alphabetic? ,@args))
  (##define-macro (char-numeric? . args)    `(##char-numeric? ,@args))
  (##define-macro (convert-case . args)     `(##convert-case ,@args))
  (##define-macro (char=? . args)           `(##char=? ,@args))
  (##define-macro (char<? . args)           `(##char<? ,@args))
  (##define-macro (cons . args)             `(##cons ,@args))
  (##define-macro (set-cdr! . args)         `(##set-cdr! ,@args))
  (##define-macro (eof-object? . args)      `(##eof-object? ,@args))
  (##define-macro (list . args)             `(##list ,@args))
  (##define-macro (make-string . args)      `(##make-string ,@args))
  (##define-macro (not . args)              `(##not ,@args))
  (##define-macro (string->number . args)   `(##string->number ,@args))
  (##define-macro (string->symbol . args)   `(##string->symbol ,@args))
  (##define-macro (string-length . args)    `(##string-length ,@args))
  (##define-macro (string-ref . args)       `(##string-ref ,@args))
  (##define-macro (string-set! . args)      `(##string-set! ,@args))
  (##define-macro (string=? . args)         `(##string=? ,@args))
  (##define-macro (substring . args)        `(##substring ,@args))
  (##define-macro (end-of-file-object? . args) `(##eof-object? ,@args))

  (##define-macro (make-vector n)           `(##make-vector ,n #f))
  (##define-macro (vector-ref . args)       `(##vector-ref ,@args))
  (##define-macro (vector-set! . args)      `(##vector-set! ,@args))
  (##define-macro (make-u8vect n)           `(##make-u8vector ,n 0))
  (##define-macro (u8vect-set! . args)      `(##u8vector-set! ,@args))
  (##define-macro (make-u16vect n)          `(##make-u16vector ,n 0))
  (##define-macro (u16vect-set! . args)     `(##u16vector-set! ,@args))
  (##define-macro (make-u32vect n)          `(##make-u32vector ,n 0))
  (##define-macro (u32vect-set! . args)     `(##u32vector-set! ,@args))
  (##define-macro (make-f32vect n)          `(##make-f32vector ,n (inexact-0)))
  (##define-macro (f32vect-set! . args)     `(##f32vector-set! ,@args))
  (##define-macro (make-f64vect n)          `(##make-f64vector ,n (inexact-0)))
  (##define-macro (f64vect-set! . args)     `(##f64vector-set! ,@args))

  (##define-macro (in-integer-range? n lo hi)
    `(and (##not (##< ,n ,lo)) (##not (##< ,hi ,n))))

  (##define-macro (sf->locat sf)                #f)
  (##define-macro (sf-peek-char sf)             `(##peek-char ,sf))
  (##define-macro (sf-read-char sf)             `(##read-char ,sf))
  (##define-macro (sf-read-error sf msg . args) `(##signal '##signal.read-error ,msg ,@args))
  (##define-macro (sf-read-error-on x msg . args) `(##signal '##signal.read-error ,msg ,@args))
  (##define-macro (make-source x locat)         x)
  (##define-macro (source-code source)          source)
  (##define-macro (source-code-set! source x)   x)
  (##define-macro (parse-keyword s)         `(##parse-keyword ,s))

  (define quote-sym            'quote)
  (define quasiquote-sym       'quasiquote)
  (define unquote-sym          'unquote)
  (define unquote-splicing-sym 'unquote-splicing)

  (define char-newline #\newline)
  (define false-object #f)
  (define (in-unicode-range? n) (##not (##< ##max-unicode n)))

  (define named-char-table   ##named-char-table)
  (define escaped-char-table ##escaped-char-table)
  (define sharp-bang-table   ##sharp-bang-table)
  (define read-table         ##read-table)

  (define (char-indexed-vector-ref tbl c)
    (let ((i (##fixnum.<-char c)))
      (if (##fixnum.< i 128)
        (##vector-ref tbl i)
        0)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; For compatibility, `read-source' is the same reader as the one used in the
; compiler.  It has been copied from the file "gambit/gsc/_source.scm".

(define (read-source sf)

  (define (read-char*)
    (let ((c (sf-read-char sf)))
      (if (end-of-file-object? c)
        (sf-read-error sf "Premature end of file encountered")
        c)))

  (define (read-non-whitespace-char)
    (let ((c (read-char*)))
      (cond ((< 0 (char-indexed-vector-ref read-table c))
             (read-non-whitespace-char))
            ((char=? c #\;)
             (let loop ()
               (if (not (char=? (read-char*) #\newline))
                 (loop)
                 (read-non-whitespace-char))))
            (else
             c))))

  (define (delimiter? c)
    (or (end-of-file-object? c)
        (not (= (char-indexed-vector-ref read-table c) 0))))

  (define (read-list first)
    (let ((result (cons first '())))
      (let loop ((end result))
        (let ((c (read-non-whitespace-char)))
          (cond ((char=? c #\)))
                ((and (char=? c #\.) (delimiter? (sf-peek-char sf)))
                 (let ((x (read-source sf)))
                   (if (char=? (read-non-whitespace-char) #\))
                     (set-cdr! end x)
                     (sf-read-error sf "')' expected"))))
                (else
                 (let ((tail (cons (rd* c) '())))
                   (set-cdr! end tail)
                   (loop tail))))))
      result))

  (define (read-blist first)
    (let ((result (cons first '())))
      (let loop ((end result))
        (let ((c (read-non-whitespace-char)))
          (cond ((char=? c #\]))
                ((and (char=? c #\.) (delimiter? (sf-peek-char sf)))
                 (let ((x (read-source sf)))
                   (if (char=? (read-non-whitespace-char) #\])
                     (set-cdr! end x)
                     (sf-read-error sf "']' expected"))))
                (else
                 (let ((tail (cons (rd* c) '())))
                   (set-cdr! end tail)
                   (loop tail))))))
      result))

  (define (read-vector kind)

    (define (loop i)
      (let ((c (read-non-whitespace-char)))
        (if (char=? c #\))
          (case kind
            ((vector)    (make-vector i))
            ((u8vector)  (make-u8vect i))
            ((u16vector) (make-u16vect i))
            ((u32vector) (make-u32vect i))
            ((f32vector) (make-f32vect i))
            ((f64vector) (make-f64vect i)))
          (let* ((x (rd* c))
                 (v (loop (+ i 1))))
            (case kind
              ((vector)
               (vector-set! v i x))
              ((u8vector)
               (let ((n (source-code x)))
                 (if (not (exact-integer-check n -128 255))
                   (sf-read-error-on x "8 bit exact integer expected"))
                 (u8vect-set! v i n)))
              ((u16vector)
               (let ((n (source-code x)))
                 (if (not (exact-integer-check n -32768 65535))
                   (sf-read-error-on x "16 bit exact integer expected"))
                 (u16vect-set! v i n)))
              ((u32vector)
               (let ((n (source-code x)))
                 (if (not (exact-integer-check n -2147483648 4294967295))
                   (sf-read-error-on x "32 bit exact integer expected"))
                 (u32vect-set! v i n)))
              ((f32vector)
               (let ((n (source-code x)))
                 (if (not (inexact-real-check n))
                   (sf-read-error-on x "inexact real expected"))
                 (f32vect-set! v i n)))
              ((f64vector)
               (let ((n (source-code x)))
                 (if (not (inexact-real-check n))
                   (sf-read-error-on x "inexact real expected"))
                 (f64vect-set! v i n))))
            v))))

    (loop 0))

  (define (read-vector* kind)
    (let ((c (read-char*)))
      (if (char=? c #\()
        (read-vector kind)
        (sf-read-error sf "'(' expected"))))

  (define (exact-integer-check n lo hi)
    (and (integer? n)
         (exact? n)
         (in-integer-range? n lo hi)))

  (define (inexact-real-check n)
    (and (real? n)
         (not (exact? n))))

  (define (char-octal? c)
    (and (not (char<? c #\0)) (not (char<? #\7 c))))

  (define (char-hexadecimal? c)
    (or (and (not (char<? c #\0)) (not (char<? #\9 c)))
        (and (not (char<? c #\a)) (not (char<? #\f c)))
        (and (not (char<? c #\A)) (not (char<? #\F c)))))

  (define (unicode n)
    (if (in-unicode-range? n)
      (unicode->character n)
      (sf-read-error sf "Character encoding out of range")))

  (define (read-escape-octal c)
    (let ((str (let loop ((i 1))
                 (let ((c (sf-peek-char sf)))
                   (if (and (< i 3)
                            (not (end-of-file-object? c))
                            (char-octal? c))
                     (begin
                       (sf-read-char sf)
                       (let ((s (loop (+ i 1))))
                         (string-set! s i c)
                         s))
                     (make-string i #\space))))))
      (string-set! str 0 c)
      (unicode (string->number str 8))))

  (define (read-escape-hexadecimal)
    (let ((c (sf-peek-char sf)))
      (if (or (end-of-file-object? c)
              (not (char-hexadecimal? c)))
        (sf-read-error sf "\\x not followed by a hexadecimal digit")
        (begin
          (sf-read-char sf)
          (let ((str (let loop ((i 1))
                       (let ((c (sf-peek-char sf)))
                         (if (and (not (end-of-file-object? c))
                                  (char-hexadecimal? c))
                           (begin
                             (sf-read-char sf)
                             (let ((s (loop (+ i 1))))
                               (string-set! s i c)
                               s))
                           (make-string i #\space))))))
            (string-set! str 0 c)
            (unicode (string->number str 16)))))))

  (define (read-escape special-escape)
    (let ((c (read-char*)))
      (cond ((char-octal? c)
             (read-escape-octal c))
            ((char=? c #\x)
             (read-escape-hexadecimal))
            ((char=? c special-escape)
             special-escape)
            (else
             (let ((x (assq c escaped-char-table)))
               (if x
                 (cdr x)
                 (sf-read-error sf "Unknown escaped character:" c)))))))

  (define (read-string closing-char)
    (define (loop i)
      (let ((c (read-char*)))
        (cond ((char=? c closing-char)
               (make-string i #\space))
              ((char=? c #\\)
               (let* ((c (read-escape closing-char))
                      (s (loop (+ i 1))))
                 (string-set! s i c)
                 s))
              (else
               (let ((s (loop (+ i 1))))
                 (string-set! s i c)
                 s)))))
    (loop 0))

  (define (read-symbol/keyword/number-string i)
    (if (delimiter? (sf-peek-char sf))
      (make-string i #\space)
      (let* ((c (sf-read-char sf))
             (s (read-symbol/keyword/number-string (+ i 1))))
        (string-set! s i (convert-case c))
        s)))

  (define (read-symbol/keyword/number c)
    (let ((s (read-symbol/keyword/number-string 1)))
      (string-set! s 0 (convert-case c))
      (or (string->number s 10)
          (parse-keyword s)
          (string->symbol s))))

  (define (read-sharp-token c)
    (let ((s (read-symbol/keyword/number-string 2)))
      (string-set! s 0 #\#)
      (string-set! s 1 (convert-case c))
      (or (string->number s 10)
          (cond ((string=? s "#f")
                 false-object)
                ((string=? s "#t")
                 #t)
                ((string=? s "#u8")
                 (read-vector* 'u8vector))
                ((string=? s "#u16")
                 (read-vector* 'u16vector))
                ((string=? s "#u32")
                 (read-vector* 'u32vector))
                ((string=? s "#f32")
                 (read-vector* 'f32vector))
                ((string=? s "#f64")
                 (read-vector* 'f64vector))
                (else
                 (sf-read-error sf "Unknown '#' token:" s))))))

  (define (read-sharp-sharp-symbol)
    (let ((s (read-symbol/keyword/number-string 2)))
      (string-set! s 0 #\#)
      (string-set! s 1 #\#)
      (string->symbol s)))

  (define (read-sharp-bang-token)
    (let ((s (read-symbol/keyword/number-string 0)))
      (string->symbol s)))

  (define (rd)
    (let ((c (sf-read-char sf)))
      (cond ((end-of-file-object? c) ;************fix
             c)
            ((< 0 (char-indexed-vector-ref read-table c))
             (rd))
            ((char=? c #\;)
             (let loop ()
               (let ((c (sf-read-char sf)))
                 (cond ((end-of-file-object? c)
                        c)
                       ((char=? c #\newline)
                        (rd))
                       (else
                        (loop))))))
            (else
             (rd* c)))))

  (define (rd* c)
    (let ((source (make-source #f (sf->locat sf))))
      (source-code-set!
        source
        (cond ((char=? c #\()
               (let ((x (read-non-whitespace-char)))
                 (if (char=? x #\))
                   '()
                   (read-list (rd* x)))))
              ((char=? c #\[)
               (let ((x (read-non-whitespace-char)))
                 (if (char=? x #\])
                   '()
                   (read-blist (rd* x)))))
              ((char=? c #\#)
               (let ((c (sf-read-char sf)))
                 (cond ((char=? c #\()
                        (read-vector 'vector))
                       ((char=? c #\\)
                        (let ((c (read-char*)))
                          (if (or (not (char-alphabetic? c))
                                  (delimiter? (sf-peek-char sf)))
                            c
                            (let ((name (read-symbol/keyword/number c)))
                              (let ((x (assq name named-char-table)))
                                (if x
                                  (cdr x)
                                  (sf-read-error sf "Unknown character name:" name)))))))
                       ((char=? c #\!)
                        (let ((name (read-sharp-bang-token)))
                          (let ((x (assq name sharp-bang-table)))
                            (if x
                              (cdr x)
                              (sf-read-error sf "Unknown '#!' token:" name)))))
                       ((char=? c #\#)
                        (read-sharp-sharp-symbol))
                       (else
                        (read-sharp-token c)))))
              ((char=? c #\")
               (read-string #\"))
              ((char=? c #\|)
               (string->symbol (read-string #\|)))
              ((char=? c #\')
               (list (make-source quote-sym (sf->locat sf))
                     (read-source sf)))
              ((char=? c #\`)
               (list (make-source quasiquote-sym (sf->locat sf))
                     (read-source sf)))
              ((char=? c #\,)
               (if (eq? (sf-peek-char sf) #\@)
                 (let ((x (make-source unquote-splicing-sym (sf->locat sf))))
                   (sf-read-char sf)
                   (list x (read-source sf)))
                 (list (make-source unquote-sym (sf->locat sf))
                       (read-source sf))))
              ((char=? c #\))
               (sf-read-error sf "Misplaced ')'"))
              ((char=? c #\])
               (sf-read-error sf "Misplaced ']'"))
              ((or (char=? c #\{)
                   (char=? c #\}))
               (sf-read-error sf "Illegal character" c))
              (else
               (if (char=? c #\.)
                 (if (delimiter? (sf-peek-char sf))
                   (sf-read-error sf "Misplaced '.'")))
               (read-symbol/keyword/number c))))))

  (rd))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  (read-source port))


(define ##named-char-table #f)
(define ##escaped-char-table #f)
(define ##sharp-bang-table #f)
(define ##read-table #f)

(define (##setup-reader-tables)

  (set! ##named-char-table
    (##list
      (##cons 'newline   #\newline)
      (##cons 'nul       (##fixnum.->char 0))
      (##cons 'backspace (##fixnum.->char 8))
      (##cons 'tab       (##fixnum.->char 9))
      (##cons 'linefeed  (##fixnum.->char 10))
      (##cons 'page      (##fixnum.->char 12))
      (##cons 'return    (##fixnum.->char 13))
      (##cons 'space     #\space)
      (##cons 'rubout    (##fixnum.->char 127))))

  (set! ##escaped-char-table
    (##list
      (##cons #\a (##fixnum.->char 7))
      (##cons #\b (##fixnum.->char 8))
      (##cons #\t (##fixnum.->char 9))
      (##cons #\n #\newline)
      (##cons #\v (##fixnum.->char 11))
      (##cons #\f (##fixnum.->char 12))
      (##cons #\r (##fixnum.->char 13))
      (##cons #\\ #\\)))

  (set! ##sharp-bang-table
    (##list
      (##cons 'eof                  #!eof)
      (##cons 'optional             #!optional)
      (##cons 'rest                 #!rest)
      (##cons 'key                  #!key)
      (##cons (##string->symbol "") #!)))

  (set! ##read-table
    (let ((rt (##make-vector 128 0)))

      ; setup whitespace chars

      (##vector-set! rt 9                           1)
      (##vector-set! rt (##fixnum.<-char #\newline) 1)
      (##vector-set! rt 12                          1)
      (##vector-set! rt 13                          1)
      (##vector-set! rt (##fixnum.<-char #\space)   1)

      ; setup other delimiters

      (##vector-set! rt (##fixnum.<-char #\;) -1)
      (##vector-set! rt (##fixnum.<-char #\() -1)
      (##vector-set! rt (##fixnum.<-char #\)) -1)
      (##vector-set! rt (##fixnum.<-char #\[) -1)
      (##vector-set! rt (##fixnum.<-char #\]) -1)
      (##vector-set! rt (##fixnum.<-char #\{) -1)
      (##vector-set! rt (##fixnum.<-char #\}) -1)
      (##vector-set! rt (##fixnum.<-char #\") -1)
      (##vector-set! rt (##fixnum.<-char #\') -1)
      (##vector-set! rt (##fixnum.<-char #\`) -1)
      (##vector-set! rt (##fixnum.<-char #\,) -1)
      (##vector-set! rt (##fixnum.<-char #\|) -1)

      rt)))

(##setup-reader-tables)

(define (##convert-case c)
  (if ##case-sensitive c (##char-downcase c)))

(define (##parse-keyword s)
  (and ##read-keywords
       (let ((len (##string-length s)))
         (and (##fixnum.< 1 len)
              (if (##eq? ##read-keywords 'prefix)
                (and (##char=? (##string-ref s 0) #\:)
                     (##string->keyword (##substring s 1 len)))
                (and (##char=? (##string-ref s (##fixnum.- len 1)) #\:)
                     (##string->keyword (##substring s 0 (##fixnum.- len 1)))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##wr-unlimited obj port display? force?)
  (##fixnum.- (max-fixnum)
              (##wr obj port display? force? (max-fixnum))))

(define (##wr-limited obj port display? force? limit)
  (##fixnum.- limit
              (##wr obj port display? force? limit)))

(define (##default-wr obj port display? force? limit)
  (if (##fixnum.< 0 limit)

    (cond ((##symbol? obj)
           (##wr-symbol obj port display? force? limit))
          ((##keyword? obj)
           (##wr-keyword obj port display? force? limit))
          ((##pair? obj)
           (##wr-pair obj port display? force? limit))
          ((##complex? obj)
           (##wr-number obj port display? force? limit))
          ((##char? obj)
           (##wr-char obj port display? force? limit))
          ((##string? obj)
           (##wr-string obj port display? force? limit))
          ((##vector? obj)
           (##wr-vector "#" (##vector->list obj) port display? force? limit))
          ((##port? obj)
           (##wr-port obj port display? force? limit))
          ((##eq? obj #t)
           (##wr-str "#t" port limit))
          ((##eq? obj #f)
           (##wr-str "#f" port limit))
          ((##eq? obj '())
           (##wr-str "()" port limit))
          ((##unbound? obj)
           (##wr-str "#<unbound>" port limit))
          ((##eq? obj (absent-obj))
           (##wr-str "#<absent>" port limit))
          ((##eq? obj (##void))
           (##wr-str "#<void>" port limit))
          ((##structure? obj)
           (##wr-structure obj port display? force? limit))
          ((##pointer? obj)
           (##wr-pointer obj port display? force? limit))
          ((##procedure? obj)
           (##wr-procedure obj port display? force? limit))
          ((##will? obj)
           (##wr-will obj port display? force? limit))
          ((##placeholder? obj)
           (##wr-placeholder obj port display? force? limit))
          ((##u8vector? obj)
           (##wr-vector "#u8"
                        (##u8vector->list obj)
                        port
                        display?
                        force?
                        limit))
          ((##u16vector? obj)
           (##wr-vector "#u16"
                        (##u16vector->list obj)
                        port
                        display?
                        force?
                        limit))
          ((##u32vector? obj)
           (##wr-vector "#u32"
                        (##u32vector->list obj)
                        port
                        display?
                        force?
                        limit))
          ((##f32vector? obj)
           (##wr-vector "#f32"
                        (##f32vector->list obj)
                        port
                        display?
                        force?
                        limit))
          ((##f64vector? obj)
           (##wr-vector "#f64"
                        (##f64vector->list obj)
                        port
                        display?
                        force?
                        limit))
          (else
           (##wr-other obj port display? force? limit)))

    0))

(define ##wr #f)
(set! ##wr ##default-wr)

(define ##wr-str
  (let ()
    (##declare (not inline)) ; it isn't worth inlining this function
    (lambda (s port limit)
      (##wr-substr s 0 (##string-length s) port limit))))

(define (##wr-substr s i j port limit)
  (let ((len (##fixnum.- j i)))
    (if (##fixnum.< limit len)
      (begin
        (##write-substring s i (##fixnum.+ i limit) port)
        0)
      (begin
        (##write-substring s i j port)
        (##fixnum.- limit len)))))

(define (##wr-ch c port limit)
  (if (##fixnum.< 0 limit)
    (begin
      (##write-char c port)
      (##fixnum.- limit 1))
    0))

(define (##wr-adr type obj port limit)
  (##wr-str ">" port
            (##wr-str (##number->string (##type-cast obj (type-fixnum)) 16) port
                      (##wr-str " #x" port
                                (##wr-str type port
                                          (##wr-str "#<" port limit))))))

(define (##wr-id-in type id name port limit)
  (##wr-str ">" port
            (##wr name port #f #f
                  (##wr-str " in " port
                            (##wr-str id port
                                      (##wr-str " " port
                                                (##wr-str type port
                                                          (##wr-str "#<" port limit))))))))

(define (##wr-named type name port limit)
  (##wr-str ">" port
            (##wr name port #f #f
                  (##wr-str " " port
                            (##wr-str type port
                                      (##wr-str "#<" port limit))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Write methods for each object type

(define (##wr-number obj port display? force? limit)
  (##wr-str (##number->string obj 10) port limit))

(define (##wr-char obj port display? force? limit)
  (if display?
    (##wr-ch obj port limit)
    (let ((x (##assq-cdr obj ##named-char-table)))
      (if x
        (##wr-str (##symbol->string (##car x)) port
                  (##wr-str "#\\" port limit))
        (##wr-ch obj port
                 (##wr-str "#\\" port limit))))))

(define (##wr-pair obj port display? force? limit)

  (define (wr-tail l limit)
    (if (##fixnum.< 0 limit)
      (let ((l (if force? (force-vars (l) l) l)))
        (cond ((##pair? l)
               (let* ((x (##car l))
                      (y (##cdr l))
                      (z (check-for-read-macro x y)))
                 (if z
                   (##wr-str ")" port
                             (##wr (##car y) port display? force?
                                   (##wr-str z port
                                             (##wr-str " . " port limit))))
                   (wr-tail y
                            (##wr x port display? force?
                                  (##wr-str " " port limit))))))
              ((##null? l)
               (##wr-str ")" port limit))
              (else
               (##wr-str ")" port
                         (##wr l port display? force?
                               (##wr-str " . " port limit))))))
      0))

  (define (wr-list x y limit)
    (wr-tail y
             (##wr x port display? force?
                   (##wr-str "(" port limit))))

  (define (check-for-read-macro head rest)
    (if (and (##pair? rest) (##null? (##cdr rest)))
      (case head
        ((quote)            "'")
        ((quasiquote)       "`")
        ((unquote)          ",")
        ((unquote-splicing) ",@")
        (else               #f))
      #f))

  (let* ((x (##car obj))
         (y (##cdr obj))
         (z (check-for-read-macro x y)))
    (if z
      (##wr (##car y) port display? force?
            (##wr-str z port limit))
      (wr-list x y limit))))

(define (##wr-procedure obj port display? force? limit)
  (let ((name (##object->global-var->identifier obj)))
    (if name
      (##wr-named "procedure" name port limit)
      (cond ((##closure? obj)
             (##wr-adr "procedure" obj port limit))
            ((##subprocedure? obj)
             (let ((parent (##object->global-var->identifier
                             (##subprocedure-parent obj))))
               (if parent
                 (##wr-id-in "subprocedure"
                             (##number->string (##subprocedure-id obj) 10)
                             parent port limit)
                 (##wr-adr "procedure" obj port limit))))
            (else
             (##wr-adr "procedure" obj port limit))))))

(define (##wr-vector prefix elems port display? force? limit)
  (##wr elems port display? force?
        (##wr-str prefix port limit)))

(define (##wr-symbol obj port display? force? limit)
  (let ((s (##symbol->string obj)))
    (if (or display? (##not (##escape-symbol? obj)))
      (##wr-str s port limit)
      (##wr-str "|" port
                (##wr-escaped-string s #\| port
                                     (##wr-str "|" port limit))))))

(define (##escape-symbol? sym)
  (let* ((s (##symbol->string sym))
         (n (##string-length s)))
    (or (##fixnum.= n 0)
        (and (##fixnum.= n 1)
             (##char=? (##string-ref s 0) #\.))
        (and (##char=? (##string-ref s 0) #\#)
             (or (##fixnum.= n 1)
                 (##not (##char=? (##string-ref s 1) #\#))))
        (##string->number s 10)
        (##parse-keyword s)
        (let loop ((i (##fixnum.- n 1)))
          (if (##fixnum.< i 0)
            #f
            (let* ((c (##string-ref s i))
                   (x (##fixnum.<-char c)))
              (or (##char=? c #\|)
                  (if (##fixnum.< x 128) ; a delimiter
                    (##not (##fixnum.= (##vector-ref ##read-table x) 0))
                    #f)
                  (##not (##char=? c (##convert-case c)))
                  (loop (##fixnum.- i 1)))))))))

(define (##wr-keyword obj port display? force? limit)
  (let ((str (##keyword->string obj)))
    (if (##eq? ##read-keywords 'prefix)
      (##wr-str str port
                (##wr-str ":" port limit))
      (##wr-str ":" port
                (##wr-str str port limit)))))

(define (##wr-port obj port display? force? limit)
  (##wr-named (if (##input-port? obj)
                (if (##output-port? obj) "input-output-port" "input-port")
                "output-port")
              (##port-name obj)
              port
              limit))

(define (##wr-structure obj port display? force? limit)
  (let* ((tag (##vector-ref obj 0))
         (n (##vector-ref tag 1)))
    (let loop ((i 1)
               (limit (##wr (##vector-ref tag 0) port display? force?
                            (##wr-str "#s(" port limit))))
      (if (##fixnum.< 0 limit)
        (if (##fixnum.< n i)
          (##wr-str ")" port limit)
          (let ((field (##vector-ref tag (##fixnum.+ i 1)))
                (value (##vector-ref obj i)))
            (let ((value (if force? (force-vars (value) value) value)))
              (loop (##fixnum.+ i 1)
                    (##wr-str ")" port
                              (##wr value port display? force?
                              (##wr-str " " port
                                        (##wr field port display? force?
                                              (##wr-str " (" port limit)))))))))
        0))))

(define (##wr-pointer obj port display? force? limit)
  (##wr-adr "pointer" obj port limit))

(define (##wr-placeholder obj port display? force? limit)
  (if force?
    (force-vars (obj)
      (##wr obj port display? force? limit))
    (##wr-adr "placeholder" obj port limit)))

(define ##escape-ctrl-chars? #f)
(set! ##escape-ctrl-chars? #t)

(define (##wr-string obj port display? force? limit)
  (if display?
    (##wr-str obj port limit)
    (##wr-str "\"" port
              (##wr-escaped-string obj #\" port
                                   (##wr-str "\"" port limit)))))

(define (##wr-escaped-string s special-escape port limit)
  (let loop ((i 0) (j 0) (limit limit))
    (if (##fixnum.< j (##string-length s))
      (let* ((c (##string-ref s j))
             (x (cond ((##char=? c special-escape) c)
                      ((##assq-cdr c ##escaped-char-table) => ##car)
                      (else #f)))
             (n (##fixnum.<-char c))
             (j+1 (##fixnum.+ j 1)))
        (if (if (##fixnum.< n 32) ##escape-ctrl-chars? x)
          (let ((new-limit
                 (##wr-substr s i j port limit)))
            (if x
              (loop j+1
                    j+1
                    (##wr-ch x port
                             (##wr-str "\\" port
                                       new-limit)))
              (let ((oct (##number->string n 8)))
                (loop j+1
                      j+1
                      (##wr-str oct port
                                (##wr-str (if (##fixnum.< n 8) "\\00" "\\0")
                                          port
                                          new-limit))))))
          (loop i j+1 limit)))
      (##wr-substr s i j port limit))))

(define (##wr-will obj port display? force? limit)
  (##wr-adr "will" obj port limit))

(define (##wr-other obj port display? force? limit)
  (let ((x (##assq-cdr obj ##sharp-bang-table)))
    (if x
      (##wr-str (##symbol->string (##car x)) port
                (##wr-str "#!" port limit))
      (##wr-str "#<unknown>" port limit))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##write obj port force?)
  (##wr-unlimited obj port #f force?))

(define (##display obj port force?)
  (##wr-unlimited obj port #t force?))

(define (##pretty obj port force? col width)

  (define (spaces n port)
    (if (##fixnum.< 0 n)
      (let ((m (if (##fixnum.< 40 n) 40 n)))
        (##write-substring "                                        " 0 m port)
        (spaces (##fixnum.- n m) port))))

  (define (indent to from port)
    (if (##fixnum.< to from)
      (begin
        (##newline port)
        (spaces to port))
      (spaces (##fixnum.- to from) port)))

  (define (obj->string obj width force?)
    (let ((port (##open-output-string)))
      (##wr-limited obj port #f force? (##fixnum.+ width 1))
      (let* ((str (##close-port port))
             (len (##string-length str)))
        (if (##fixnum.< width len) #f str))))

  (define (p obj port force? col width extra pp-pair)
    (let ((obj (if force? (force-vars (obj) obj) obj)))
      (if (or (##pair? obj)
              (##vector? obj)
              (##u8vector? obj)
              (##u16vector? obj)
              (##u32vector? obj)
              (##f32vector? obj)
              (##f64vector? obj)
              (##structure? obj))
        (let ((str
               (obj->string
                 obj
                 (##fixnum.- (##fixnum.- width col) extra) force?)))
          (if str
            (begin
              (##write-string str port)
              (##fixnum.+ col (##string-length str)))
            (cond ((##pair? obj)
                   (pp-pair obj port force? col width extra))
                  ((##vector? obj)
                   (let ((col* (##fixnum.+ col 1))
                         (elems (##vector->list obj)))
                     (##write-string "#" port)
                     (pp-list elems port force? col* width extra pp-expr)))
                  ((##u8vector? obj)
                   (let ((col* (##fixnum.+ col 3))
                         (elems (##u8vector->list obj)))
                     (##write-string "#u8" port)
                     (pp-list elems port force? col* width extra pp-expr)))
                  ((##u16vector? obj)
                   (let ((col* (##fixnum.+ col 4))
                         (elems (##u16vector->list obj)))
                     (##write-string "#u16" port)
                     (pp-list elems port force? col* width extra pp-expr)))
                  ((##u32vector? obj)
                   (let ((col* (##fixnum.+ col 4))
                         (elems (##u32vector->list obj)))
                     (##write-string "#u32" port)
                     (pp-list elems port force? col* width extra pp-expr)))
                  ((##f32vector? obj)
                   (let ((col* (##fixnum.+ col 4))
                         (elems (##f32vector->list obj)))
                     (##write-string "#f32" port)
                     (pp-list elems port force? col* width extra pp-expr)))
                  ((##f64vector? obj)
                   (let ((col* (##fixnum.+ col 4))
                         (elems (##f64vector->list obj)))
                     (##write-string "#f64" port)
                     (pp-list elems port force? col* width extra pp-expr)))
                  (else
                   (let ((col* (##fixnum.+ col 2))
                         (elems
                           (let* ((tag (##vector-ref obj 0))
                                  (n (##vector-ref tag 1)))
                             (let loop ((i n) (l '()))
                               (if (##fixnum.< 0 i)
                                 (let ((field
                                        (##vector-ref tag (##fixnum.+ i 1)))
                                       (value
                                        (##vector-ref obj i)))
                                   (loop (##fixnum.- i 1)
                                         (##cons (##list field value) l)))
                                 (##cons (##vector-ref tag 0) l))))))
                     (##write-string "#s" port)
                     (pp-pair elems port force? col* width extra))))))
        (##fixnum.+ col (##write obj port force?)))))

  (define (pp-expr expr port force? col width extra)
    (let ((head (##car expr)))
      (let* ((head (if force? (force-vars (head) head) head))
             (style (pp-style head)))
        (if style
          (style expr port force? col width extra)
          (if (##symbol? head)
            (if (##fixnum.< (##string-length (##symbol->string head)) 8)
              (pp-call expr port force? col width extra pp-expr)
              (pp-general expr port force? col width extra #f #f #f pp-expr))
            (pp-list expr port force? col width extra pp-expr))))))

  ; (head item1
  ;       item2
  ;       item3)
  (define (pp-call expr port force? col width extra pp-item)
    (##write-string "(" port)
    (let* ((head (##car expr))
           (rest (##cdr expr))
           (col* (##fixnum.+ (##fixnum.+ col 1) (##write head port force?))))
      (pp-down rest port force? col* (##fixnum.+ col* 1) width extra pp-item)))

  ; (item1
  ;  item2
  ;  item3)
  (define (pp-list l port force? col width extra pp-item)
    (##write-string "(" port)
    (let ((col* (##fixnum.+ col 1)))
      (pp-down l port force? col* col* width extra pp-item)))

  (define (pp-down l port force? col1 col2 width extra pp-item)
    (let loop ((l l) (col* col1))
      (if (##pair? l)
        (let ((rest (##cdr l)))
          (let* ((rest (if force? (force-vars (rest) rest) rest))
                 (extra* (if (##null? rest) (##fixnum.+ extra 1) 0)))
            (indent col2 col* port)
            (loop rest (p (##car l) port force? col2 width extra* pp-item))))
        (if (##null? l)
          (begin
            (##write-string ")" port)
            (##fixnum.+ col* 1))
          (begin
            (indent col2 col* port)
            (##write-string "." port)
            (indent col2 col* port)
            (let* ((extra* (##fixnum.+ extra 1))
                   (col** (p l port force? col2 width extra* pp-item)))
              (##write-string ")" port)
              (##fixnum.+ col** 1)))))))

  (define (pp-expr-list l port force? col width extra)
    (pp-list l port force? col width extra pp-expr))

  (define (pp-abbrev expr port force? col width extra prefix)
    (let* ((rest (##cdr expr))
           (rest (if force? (force-vars (rest) rest) rest)))
      (if (and (##pair? rest) (##null? (##cdr rest)))
        (let ((col* (##fixnum.+ col (##string-length prefix))))
          (##write-string prefix port)
          (p (##car rest) port force? col* width extra pp-expr))
        (pp-call expr port force? col width extra pp-expr))))

  (define (pp-general expr port force? col width extra named? pp-1 pp-2 pp-3)

    (define (tail1 rest col1 col2 col3)
      (if (and pp-1 (##pair? rest))
        (begin
          (indent col3 col2 port)
          (let* ((val1 (##car rest))
                 (rest (##cdr rest))
                 (rest (if force? (force-vars (rest) rest) rest))
                 (extra* (if (##null? rest) (##fixnum.+ extra 1) 0))
                 (col* (p val1 port force? col3 width extra* pp-1)))
            (tail2 rest col1 col* col3)))
        (tail2 rest col1 col2 col3)))

    (define (tail2 rest col1 col2 col3)
      (if (and pp-2 (##pair? rest))
        (begin
          (indent col3 col2 port)
          (let* ((val1 (##car rest))
                 (rest (##cdr rest))
                 (rest (if force? (force-vars (rest) rest) rest))
                 (extra* (if (##null? rest) (##fixnum.+ extra 1) 0))
                 (col* (p val1 port force? col3 width extra* pp-2)))
            (tail3 rest col1 col*)))
        (tail3 rest col1 col2)))

    (define (tail3 rest col1 col2)
      (pp-down rest port force? col2 col1 width extra pp-3))

    (##write-string "(" port)
    (let* ((head (##car expr))
           (rest (##cdr expr))
           (rest (if force? (force-vars (rest) rest) rest))
           (col* (##fixnum.+ (##fixnum.+ col 1) (##write head port force?))))
      (if (and named? (##pair? rest))
        (begin
          (##write-string " " port)
          (let* ((name (##car rest))
                 (rest (##cdr rest))
                 (rest (if force? (force-vars (rest) rest) rest))
                 (col** (##fixnum.+ (##fixnum.+ col* 1) (##write name port force?))))
            (tail1 rest (##fixnum.+ col 2) col** (##fixnum.+ col** 1))))
        (tail1 rest (##fixnum.+ col 2) col* (##fixnum.+ col* 1)))))

  (define (pp-quote expr port force? col width extra)
    (pp-abbrev expr port force? col width extra "'"))

  (define (pp-quasiquote expr port force? col width extra)
    (pp-abbrev expr port force? col width extra "`"))

  (define (pp-unquote expr port force? col width extra)
    (pp-abbrev expr port force? col width extra ","))

  (define (pp-unquote-splicing expr port force? col width extra)
    (pp-abbrev expr port force? col width extra ",@"))

  (define (pp-lambda expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr-list #f pp-expr))

  (define (pp-if expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr #f pp-expr))

  (define (pp-set! expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr #f pp-expr))

  (define (pp-cond expr port force? col width extra)
    (pp-call expr port force? col width extra pp-expr-list))

  (define (pp-case expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr #f pp-expr-list))

  (define (pp-and expr port force? col width extra)
    (pp-call expr port force? col width extra pp-expr))

  (define (pp-or expr port force? col width extra)
    (pp-call expr port force? col width extra pp-expr))

  (define (pp-let expr port force? col width extra)
    (let* ((rest (##cdr expr))
           (rest (if force? (force-vars (rest) rest) rest))
           (named? (and (##pair? rest) (##symbol? (##car rest)))))
      (pp-general expr port force? col width extra named? pp-expr-list #f pp-expr)))

  (define (pp-let* expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr-list #f pp-expr))

  (define (pp-letrec expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr-list #f pp-expr))

  (define (pp-begin expr port force? col width extra)
    (pp-general expr port force? col width extra #f #f #f pp-expr))

  (define (pp-do expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr-list pp-expr-list pp-expr))

  (define (pp-define expr port force? col width extra)
    (pp-general expr port force? col width extra #f pp-expr-list #f pp-expr))

  (define (pp-style x)
    (case x
      ((quote) pp-quote)
      ((quasiquote) pp-quasiquote)
      ((unquote) pp-unquote)
      ((unquote-splicing) pp-unquote-splicing)
      ((lambda) pp-lambda)
      ((if) pp-if)
      ((set!) pp-set!)
      ((cond) pp-cond)
      ((case) pp-case)
      ((and) pp-and)
      ((or) pp-or)
      ((let) pp-let)
      ((let*) pp-let*)
      ((letrec) pp-letrec)
      ((begin) pp-begin)
      ((do) pp-do)
      ((define) pp-define)
      (else #f)))

  (p obj port force? col width 0 pp-expr))

(define (##pretty-print obj port width)
  (##pretty obj port (if-forces #t #f) 0 width)
  (##newline port))

(define (##object->string obj width force?)
  (let ((port (##open-output-string)))
    (##wr-limited obj port #f force? (##fixnum.+ width 1))
    (let* ((str (##close-port port))
           (len (##string-length str)))
      (if (##fixnum.< width len)
        (begin
          (##string-set! str (##fixnum.- width 1) #\.)
          (##string-set! str (##fixnum.- width 2) #\.)
          (##string-set! str (##fixnum.- width 3) #\.)
          (##string-shrink! str width)
          str)
        str))))

;(define (##format port str . args)
;  (let ((len (##string-length str)))
;    (let loop ((i 0) (j 0) (args args))
;      (if (##not (##fixnum.< j len))
;        (##write-substring str i j port)
;        (let ((c (##string-ref str j)))
;          (if (##char=? c #\~)
;            (let ((c (##string-ref str (##fixnum.+ j 1))))
;              (##write-substring str i j port)
;              (if (##memq c '(#\A #\S #\V #\D #\B #\O #\X))
;                (let ((arg (##car args))
;                      (rest (##cdr args)))
;                  (cond ((##char=? c #\A)
;                         (##display arg port #t))
;                        ((##char=? c #\S)
;                         (##write arg port #t))
;                        ((##char=? c #\V)
;                         (##wr-unlimited arg port #f #f))
;                        ((##char=? c #\D)
;                         (##write-string (##number->string arg 10) port))
;                        ((##char=? c #\B)
;                         (##write-string (##number->string arg 2) port))
;                        ((##char=? c #\O)
;                         (##write-string (##number->string arg 8) port))
;                        ((##char=? c #\X)
;                         (##write-string (##number->string arg 16) port)))
;                  (loop (##fixnum.+ j 2) (##fixnum.+ j 2) rest))
;                (cond ((##char=? c #\%)
;                       (##newline port)
;                       (loop (##fixnum.+ j 2) (##fixnum.+ j 2) args))
;                      ((##char=? c #\~)
;                       (##write-string "~" port)
;                       (loop (##fixnum.+ j 2) (##fixnum.+ j 2) args))
;                      ((##char=? c #\newline)
;                       (let ((k (let skip ((j (##fixnum.+ j 2)))
;                                  (cond ((##not (##fixnum.< j len))
;                                         j)
;                                        ((##char-whitespace? c)
;                                         (skip (##fixnum.+ j 1)))
;                                        (else
;                                         j)))))
;                         (loop k k args)))
;                      (else
;                       (loop (##fixnum.+ j 2) (##fixnum.+ j 2) args)))))
;            (loop i (##fixnum.+ j 1) args)))))))

;------------------------------------------------------------------------------

(define (##current-input-port)
  (##dynamic-ref '##current-input-port ##stdin))

(define (##current-output-port)
  (##dynamic-ref '##current-output-port ##stdout))

;------------------------------------------------------------------------------

(##declare (not interrupts-enabled))

; IEEE Scheme procedures:

(define (call-with-input-file
         path
         proc
         #!optional (char-encoding (absent-obj)))
  (force-vars (path proc char-encoding)
    (check-string path (call-with-input-file path proc char-encoding)
      (check-procedure proc (call-with-input-file path proc char-encoding)
        (let ((port
               (##open-input-file
                path
                (if (##eq? char-encoding (absent-obj)) #f char-encoding))))
          (if port
            (let ((result (proc port)))
              (##close-port port)
              result)
            (trap-open-file
             (call-with-input-file path proc char-encoding))))))))

(define (call-with-output-file
         path
         proc
         #!optional (char-encoding (absent-obj)))
  (force-vars (path proc char-encoding)
    (check-string path (call-with-output-file path proc char-encoding)
      (check-procedure proc (call-with-output-file path proc char-encoding)
        (let ((port
               (##open-output-file
                path
                (if (##eq? char-encoding (absent-obj)) #f char-encoding))))
          (if port
            (let ((result (proc port)))
              (##close-port port)
              result)
            (trap-open-file
             (call-with-output-file path proc char-encoding))))))))

(define (input-port? x)
  (force-vars (x)
    (and (##port? x) (##input-port? x))))

(define (output-port? x)
  (force-vars (x)
    (and (##port? x) (##output-port? x))))

(define (current-input-port)
  (##current-input-port))

(define (current-output-port)
  (##current-output-port))

(define (open-input-file
         path
         #!optional (char-encoding (absent-obj)))
  (force-vars (path char-encoding)
    (check-string path (open-input-file path char-encoding)
      (let ((port
             (##open-input-file
              path
              (if (##eq? char-encoding (absent-obj)) #f char-encoding))))
        (if port
          port
          (trap-open-file (open-input-file path char-encoding)))))))

(define (open-output-file
         path
         #!optional (char-encoding (absent-obj)))
  (force-vars (path char-encoding)
    (check-string path (open-output-file path char-encoding)
      (let ((port
             (##open-output-file
              path
              (if (##eq? char-encoding (absent-obj)) #f char-encoding))))
        (if port
          port
          (trap-open-file (open-output-file path char-encoding)))))))

(define (close-input-port p)
  (force-vars (p)
    (check-input-port p (close-input-port p)
      (##close-port p))))

(define (close-output-port p)
  (force-vars (p)
    (check-output-port p (close-output-port p)
      (##close-port p))))

(define-system (##eof-object? x)
  (##eq? x #!eof))

(define (eof-object? x)
  (force-vars (x)
    (##eof-object? x)))

(define (read #!optional (p (absent-obj)))
  (force-vars (p)
    (let ((port
           (if (##eq? p (absent-obj))
             (##current-input-port)
             p)))
      (check-input-port port (read p)
        (check-open-port port (read p)
          (##read port))))))

(define (read-char #!optional (p (absent-obj)))
  (force-vars (p)
    (let ((port
           (if (##eq? p (absent-obj))
             (##current-input-port)
             p)))
      (check-input-port port (read-char p)
        (check-open-port port (read-char p)
          (##read-char port))))))

(define (peek-char #!optional (p (absent-obj)))
  (force-vars (p)
    (let ((port
           (if (##eq? p (absent-obj))
             (##current-input-port)
             p)))
      (check-input-port port (peek-char p)
        (check-open-port port (peek-char p)
          (##peek-char port))))))
  
(define (write obj #!optional (p (absent-obj)))
  (force-vars (obj p)
    (let ((port
           (if (##eq? p (absent-obj))
             (##current-output-port)
             p)))
      (check-output-port port (write obj p)
        (check-open-port port (write obj p)
          (begin
            (##write obj port (if-forces #t #f))
            (##void)))))))

(define (display obj #!optional (p (absent-obj)))
  (force-vars (obj p)
    (let ((port
           (if (##eq? p (absent-obj))
             (##current-output-port)
             p)))
      (check-output-port port (display obj p)
        (check-open-port port (display obj p)
          (begin
            (##display obj port (if-forces #t #f))
            (##void)))))))

(define (newline #!optional (p (absent-obj)))
  (force-vars (p)
    (let ((port
           (if (##eq? p (absent-obj))
             (##current-output-port)
             p)))
      (check-output-port port (newline p)
        (check-open-port port (newline p)
          (begin
            (##newline port)
            (##void)))))))

(define (write-char c #!optional (p (absent-obj)))
  (force-vars (c p)
    (let ((port
           (if (##eq? p (absent-obj))
             (##current-output-port)
             p)))
      (check-char c (write-char c p)
        (check-output-port port (write-char c p)
          (check-open-port port (write-char c p)
            (begin
              (##write-char c port)
              (##void))))))))

;------------------------------------------------------------------------------

; R4RS Scheme procedures:

(define (with-input-from-file
         path
         thunk
         #!optional (char-encoding (absent-obj)))
  (force-vars (path thunk char-encoding)
    (check-string path (with-input-from-file path thunk char-encoding)
      (check-procedure thunk (with-input-from-file path thunk char-encoding)
        (let ((port
               (##open-input-file
                path
                (if (##eq? char-encoding (absent-obj)) #f char-encoding))))
          (if port
            (let ((result
                    (##dynamic-let
                      (##list (##cons '##current-input-port port))
                      thunk)))
              (##close-port port)
              result)
            (trap-open-file (with-input-from-file path thunk char-encoding))))))))

(define (with-output-to-file
         path
         thunk
         #!optional (char-encoding (absent-obj)))
  (force-vars (path thunk char-encoding)
    (check-string path (with-output-to-file path thunk char-encoding)
      (check-procedure thunk (with-output-to-file path thunk char-encoding)
        (let ((port
               (##open-output-file
                path
                (if (##eq? char-encoding (absent-obj)) #f char-encoding))))
          (if port
            (let ((result
                    (##dynamic-let
                      (##list (##cons '##current-output-port port))
                      thunk)))
              (##close-port port)
              result)
            (trap-open-file (with-output-to-file path thunk char-encoding))))))))

(define (char-ready? #!optional (p (absent-obj)))
  (force-vars (p)
    (let ((port
           (if (##eq? p (absent-obj))
             (##current-input-port)
             p)))
      (check-input-port port (char-ready? p)
        (check-open-port port (char-ready? p)
          (##char-ready? port))))))

(define (transcript-on path)
  (##void))

(define (transcript-off)
  (##void))

;------------------------------------------------------------------------------

; Non-standard procedures:

(define (file-exists? path)
  (force-vars (path)
    (check-string path (file-exists? path)
      (let ((port (##open-input-file path #f)))
        (if port
          (begin
            (##close-port port)
            #t)
          #f)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (flush-output #!optional (p (absent-obj)))
  (force-vars (p)
    (let ((port
           (if (##eq? p (absent-obj))
             (##current-output-port)
             p)))
      (check-output-port port (flush-output p)
        (check-open-port port (flush-output p)
          (begin
            (##flush-output port)
            (##void)))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (open-input-string s)
  (force-vars (s)
    (check-string s (open-input-string s)
      (##open-input-string s))))

(define (open-output-string)
  (##open-output-string))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (with-input-from-string s thunk)
  (force-vars (s thunk)
    (check-string s (with-input-from-string s thunk)
      (check-procedure thunk (with-input-from-string s thunk)
        (let ((port (##open-input-string s)))
          (##dynamic-let
            (##list (##cons '##current-input-port port))
            thunk))))))

(define (with-output-to-string thunk)
  (force-vars (thunk)
    (check-procedure thunk (with-output-to-string thunk)
      (let ((port (##open-output-string)))
        (##dynamic-let
          (##list (##cons '##current-output-port port))
          thunk)
        (##close-port port)))))

(define (call-with-input-string s proc)
  (force-vars (s proc)
    (check-string s (call-with-input-string s proc)
      (check-procedure proc (call-with-input-string s proc)
        (let ((port (##open-input-string s)))
          (proc port))))))

(define (call-with-output-string proc)
  (force-vars (proc)
    (check-procedure proc (call-with-output-string proc)
      (let ((port (##open-output-string)))
        (proc port)
        (##close-port port)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (with-input-from-port port thunk)
  (force-vars (port thunk)
    (check-input-port port (with-input-from-port port thunk)
      (check-open-port port (with-input-from-port port thunk)
        (check-procedure thunk (with-input-from-port port thunk)
          (##dynamic-let
            (##list (##cons '##current-input-port port))
            thunk))))))

(define (with-output-to-port port thunk)
  (force-vars (port thunk)
    (check-output-port port (with-output-to-port port thunk)
      (check-open-port port (with-output-to-port port thunk)
        (check-procedure thunk (with-output-to-port port thunk)
          (##dynamic-let
            (##list (##cons '##current-output-port port))
            thunk))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (pretty-print obj #!optional (p (absent-obj)) (w (absent-obj)))
  (force-vars (obj p w)
    (let ((port
           (if (##eq? p (absent-obj))
             (##current-output-port)
             p)))
      (check-output-port port (pp obj p w)
        (check-open-port port (pp obj p w)
          (let ((width
                 (if (##eq? w (absent-obj))
                   (##port-width port)
                   w)))
            (check-exact-int-non-neg width (pp obj p w)
              (begin
                (##pretty-print obj port width)
                (##void)))))))))

(define (pp obj #!optional (p (absent-obj)) (w (absent-obj)))
  (force-vars (obj p w)
    (let ((port
           (if (##eq? p (absent-obj))
             (##current-output-port)
             p)))
      (check-output-port port (pp obj p w)
        (check-open-port port (pp obj p w)
          (let ((width
                 (if (##eq? w (absent-obj))
                   (##port-width port)
                   w)))
            (check-exact-int-non-neg width (pp obj p w)
              (begin
                (if (##procedure? obj)
                  (##pretty-print (##decompile obj) port width)
                  (##pretty-print obj port width))
                (##void)))))))))

;------------------------------------------------------------------------------
