;==============================================================================

; file: "_nonstd.scm"

; Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

;------------------------------------------------------------------------------

; Non-standard procedures

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(##define-macro (define-runtime-macro pattern . rest)

  (define (form-size parms) ; this definition must match the one in "_eval.scm"
    (let loop ((l parms) (n 1))
      (if (pair? l)
        (loop (cdr l) (+ n 1))
        (if (null? l) n (- 0 n)))))

  `(##top-cte-add-macro!
     ##interaction-cte
     ',(car pattern)
     (##cons ',(form-size (cdr pattern))
             (lambda ,(cdr pattern) ,@rest))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (exit #!optional (status (absent-obj)))
  (if (##eq? status (absent-obj))
    (##exit)
    (force-vars (status)
      (##exit status))))

(define (error msg . args)
  (##call-with-current-continuation (lambda (cont) (##sequentially (lambda ()
    (##identify-error #f msg args '())
    (##debug-repl cont))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (define-macro pattern body . rest)
  `(##define-macro ,pattern ,body ,@rest))

(define-runtime-macro (declare . rest)
  `(##declare ,@rest))

(define-runtime-macro (include filename)
  `(##include ,filename))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (dynamic-define var val)
  `(##dynamic-define ',var ,val))

(define-runtime-macro (dynamic-ref var)
  `(##dynamic-ref ',var))

(define-runtime-macro (dynamic-set! var val)
  `(##dynamic-set! ',var ,val))

(define-runtime-macro (dynamic-let bindings body . rest)
  (##dynamic-let-build (##cons bindings (##cons body rest))))

(define (##dynamic-let-build src)

  (define (build src bindings)
    (if (##pair? bindings)
      (let ((binding (##force-list (##car bindings))))
        (##shape src binding 2)
        (let ((x (##car binding)))
          (force-vars (x)
            (let ((rest (build src (##cdr bindings))))
              (if (##not (##symbol? x))
                (##signal '##signal.syntax-error src "Identifier expected:" x))
              (##cons (##list '##cons (##list 'quote x) (##cadr binding))
                      rest)))))
      (if (##null? bindings)
        '()
        (##signal '##signal.syntax-error src "Ill-terminated bindings"))))

  (##list '##dynamic-let
          (##cons '##list (build src (##car src)))
          (##cons 'lambda (##cons '() (##cdr src)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (define-structure name . field-defs)

  (define (err)
    (##signal '##signal.syntax-error
              (##cons 'define-structure (##cons name field-defs))
              "Ill-formed special form:"
              'define-structure))

  (define (sym . strings)
    (##string->symbol (##apply ##string-append strings)))

  (if (##symbol? name)
    (let ((name-str (##symbol->string name)))

      (define (generate-accessors fields i tag)
        (if (##null? fields)
          '()
          (let* ((field (##car fields))
                 (field-str (##symbol->string field))
                 (field-ref (sym name-str "-" field-str))
                 (field-set! (sym name-str "-" field-str "-set!")))
            (##cons `(define ,field-set!
                       (lambda (x y)
                         (##declare (extended-bindings))
                         (##structure-set! x ,i y ',tag)))
                    (##cons `(define ,field-ref
                               (lambda (x)
                                 (##declare (extended-bindings))
                                 (##structure-ref x ,i ',tag)))
                            (generate-accessors (##cdr fields)
                                                (##fixnum.+ i 1)
                                                tag))))))

      (define (generate-definitions rev-all-fields rev-printed-fields)
        (let* ((all-fields (##reverse rev-all-fields))
               (printed-fields (##reverse rev-printed-fields))
               (tag (##list->vector
                      (##cons name
                              (##cons (##length printed-fields)
                                      all-fields)))))
          `(begin
             ,@(generate-accessors all-fields 1 tag)
             (define ,(sym name-str "?")
               (lambda (x)
                 (##declare (extended-bindings))
                 (and (##structure? x)
                      (##eq? (##vector-ref x 0) ',tag))))
             (define ,(sym "make-" name-str)
               (lambda (,@all-fields)
                 (##declare (extended-bindings))
                 (##subtype-set! (##vector ',tag ,@all-fields)
                                 ,(subtype-structure)))))))

      (let loop1 ((l1 field-defs) (l2 '()))
        (if (##pair? l1)
          (let ((rest (##cdr l1)) (field (##car l1)))
            (cond ((##symbol? field)
                   (loop1 rest (##cons field l2)))
                  ((and (or (##null? field) (##pair? field))
                        (##not (##pair? rest)))
                   (let ((printed-fields l2))
                     (let loop2 ((l1 field) (l2 l2))
                       (if (##pair? l1)
                         (let ((rest (##cdr l1)) (field (##car l1)))
                           (cond ((##symbol? field)
                                  (loop2 rest (##cons field l2)))
                                  (else
                                   (err))))
                         (generate-definitions l2 printed-fields)))))
                  (else
                   (err))))
          (generate-definitions l2 l2))))
    (err)))

(define (##structure-ref obj i tag)
  (if (and (##structure? obj)
           (##eq? (##vector-ref obj 0) tag))
    (##vector-ref obj i)
    (##structure-error tag i #f (##list obj))))

(define (##structure-set! obj i v tag)
  (if (and (##structure? obj)
           (##eq? (##vector-ref obj 0) tag))
    (begin (##vector-set! obj i v) (##void))
    (##structure-error tag i #t (##list obj v))))

(define (##structure-error tag i set!? args)
  (let ((name-str (##symbol->string (##vector-ref tag 0)))
        (field-str (##symbol->string (##vector-ref tag (##fixnum.+ i 1)))))
    (##runtime-error (##string-append "Structure of type `"
                                      name-str
                                      "' expected")
                     (##string->symbol
                       (if set!?
                         (##string-append name-str "-" field-str "-set!")
                         (##string-append name-str "-" field-str)))
                     args)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (trace . procs)

  (define (tr l)
    (if (##pair? l)
      (let ((var (##car l)))
        (##cons (##list '##trace
                        (##list 'quote var)
                        (##list 'lambda '() var)
                        (##list 'lambda '(##val) (##list 'set! var '##val)))
                (tr (##cdr l))))
      '()))

  (if (##pair? procs)
    (##cons 'begin (tr procs))
    '(##trace-list)))

(define-runtime-macro (untrace . procs)

  (define (untr l)
    (if (##pair? l)
      (let ((var (##car l)))
        (##cons (##list '##untrace (##list 'quote var)) (untr (##cdr l))))
      '()))

  (if (##pair? procs)
    (##cons 'begin (untr procs))
    '(##untrace-all)))

(define ##traced '())

(define (##trace name getter setter)

  (define (add-quotes l)
    (if (##pair? l)
      (let ((x (##car l)))
        (##cons (if (##self-eval? x) x (##list 'quote x))
                (add-quotes (##cdr l))))
      '()))

  (define indent-limit 10)

  (define (traced-proc proc)
    (lambda args
      (let ((out (##repl-out)))

        (define (bars i)
          (let loop ((j 0))
            (if (##fixnum.< j i)
              (begin
                (##write-string
                  (if (##fixnum.= (##fixnum.remainder j 2) 0) "|" " ")
                  out)
                (loop (##fixnum.+ j 1))))))

        (define (indent i)
          (if (##fixnum.< i indent-limit)
            (bars (##fixnum.+ i 1))
            (let ((i-str (##number->string i 10)))
              (bars (##fixnum.- indent-limit
                                (##fixnum.+ (##string-length i-str) 1)))
              (##write-string "[" out)
              (##write-string i-str out)
              (##write-string "]" out))))

        (define (indent-width i)
          (if (##fixnum.< i indent-limit)
            (##fixnum.+ i 1)
            (##fixnum.+ indent-limit 1)))

        (let* ((i (##dynamic-ref '##trace-indent 0))
               (w (##fixnum.- (##port-width out) (indent-width i))))

          (indent i)
          (##write-string
            (##object->string
              (##cons name (add-quotes args))
              w
              (if-forces #t #f))
            out)
          (##newline out)
          (let ((result
                  (##dynamic-let
                    (##list (##cons '##trace-indent (##fixnum.+ i 1)))
                    (lambda () (##apply proc args)))))
            (indent i)
            (##write-string
              (##object->string result w (if-forces #t #f))
              out)
            (##newline out)
            result)))))

  (let ((proc (getter)))
    (if (##procedure? proc)
      (let ((x (##assq name ##traced)))
        (if (##not (and x (##eq? proc (##cadddr x)))) ; being traced already?
          (let ((tproc (traced-proc proc)))
            (if x
              (begin
                (if (##eq? ((##cadr x)) (##cadddr x)) ; var = traced proc?
                  ((##caddr x) (##car (##cddddr x)))) ; restore old value
                (##set-car! (##cdr x) getter)
                (##set-car! (##cddr x) setter)
                (##set-car! (##cdddr x) tproc)
                (##set-car! (##cddddr x) proc))
              (set! ##traced
                (##cons (##list name getter setter tproc proc) ##traced)))
            (setter tproc))))))

  (##void))

(define (##trace-list)
  (let loop ((l1 ##traced) (l2 '()))
    (if (##pair? l1)
      (let ((x (##car l1)))
        (loop (##cdr l1) (##cons (##car x) l2)))
      l2)))

(define (##untrace name)
  (let loop ((l1 ##traced) (l2 '()))
    (if (##pair? l1)
      (let ((x (##car l1)))
        (if (##eq? (##car x) name)
          (begin
            (if (##eq? ((##cadr x)) (##cadddr x)) ; var = traced proc?
              ((##caddr x) (##car (##cddddr x)))) ; restore old value
            (set! ##traced (##append (##reverse l2) (##cdr l1)))
            (##void))
          (loop (##cdr l1) (##cons x l2))))
      (##void))))

(define (##untrace-all)
  (let loop ((l ##traced))
    (if (##pair? l)
      (let ((x (##car l)))
        (##untrace (##car x))
        (loop (##cdr l)))
      (##void))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (set-gc-report report?)
  (set! ##gc-report report?)
  (##void))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (void)
  (##void))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (runtime)
  (##/ (##exact->inexact (##process-time)) 1000))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (time expr)
  `(##time (lambda () ,expr) ',expr))

(define (##time thunk expr)
  (let ((start (##process-time)))
    (let ((result (thunk)))
      (let ((stop (##process-time)))
        (let ((out (##repl-out)))
          (##write (##list 'time expr) out #f)
          (##newline out)
          (##write-string "    " out)
          (##write (##- stop start) out #f)
          (##write-string " ms elapsed cpu time (user+system time)" out)
          (##newline out)
          result)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##gensym-count -1)

(define (gensym #!optional (p (absent-obj)))
  (force-vars (p)
    (let ((prefix (if (##eq? p (absent-obj)) 'g p)))
      (check-symbol prefix (gensym p)
        (##make-uninterned-symbol
          (##string-append (##symbol->string prefix)
                           (begin
                             (set! ##gensym-count (##+ ##gensym-count 1))
                             (##number->string ##gensym-count 10))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (will? x)
  (force-vars (x)
    (##will? x)))

(define (make-will owner agent)
  (force-vars (agent)
    (check-procedure agent (make-will owner agent)
      (##make-will owner agent))))

(define (will-owner x)
  (force-vars (x)
    (check-will x (will-owner x)
      (##will-owner x))))

;------------------------------------------------------------------------------
