;==============================================================================

; file: "_repl.scm"

; Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

;==============================================================================

; Decompilation of a piece of code

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(##define-macro (mk-degen params . def)
  `(let () (##declare (not inline)) (lambda ($code ,@params) ,@def)))

(##define-macro (degen proc . args)
  `(,proc $code ,@args))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##extract-container $code rte)
  (let loop ((c (code-cte $code)) (r rte))
    (cond ((##cte-top? c)
           #f)
          ((##cte-frame? c)
           (let ((vars (##cte-frame-vars c)))
             (if (and (##pair? vars) (##eq? (##car vars) (self-var)))
               (rte-ref r 1)
               (loop (##cte-parent-cte c) (rte-up r)))))
          (else
           (loop (##cte-parent-cte c) r)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##begin? x) (and (##pair? x) (##eq? (##car x) 'begin)))
(define (##cond? x)  (and (##pair? x) (##eq? (##car x) 'cond)))
(define (##and? x)   (and (##pair? x) (##eq? (##car x) 'and)))
(define (##or? x)    (and (##pair? x) (##eq? (##car x) 'or)))
(define (##void-constant? x)
  (and (##pair? x)
       (##eq? (##car x) 'quote)
       (##eq? (##cadr x) (##void))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##degen-top
  (mk-degen ()
    (##decomp (^ 0))))

(define ##degen-cst-x
  (mk-degen (val)
    (if (##self-eval? val) val (##list 'quote val))))

(define ##degen-cst
  (mk-degen ()
    (degen ##degen-cst-x (^ 0))))

(define ##degen-loc-ref-x-y
  (mk-degen (up over)
    (degen ##degen-up-over up over)))

(define ##degen-up-over
  (mk-degen (up over)
    (let loop1 ((c (code-cte $code)) (up up))
      (cond ((##cte-frame? c)
             (if (##fixnum.= up 0)
               (let loop2 ((vars (##cte-frame-vars c)) (i over))
                 (if (##fixnum.< i 2)
                   (##car vars)
                   (loop2 (##cdr vars) (##fixnum.- i 1))))
               (loop1 (##cte-parent-cte c) (##fixnum.- up 1))))
            (else
             (loop1 (##cte-parent-cte c) up))))))

(define ##degen-loc-ref
  (mk-degen ()
    (degen ##degen-loc-ref-x-y (^ 0) (^ 1))))

(define ##degen-glo-ref
  (mk-degen ()
    (global-var->identifier (^ 0))))

(define ##degen-loc-set
  (mk-degen ()
    (##list 'set! (degen ##degen-up-over (^ 1) (^ 2))
                  (##decomp (^ 0)))))

(define ##degen-glo-set
  (mk-degen ()
    (##list 'set! (global-var->identifier (^ 1))
                  (##decomp (^ 0)))))

(define ##degen-glo-def
  (mk-degen ()
    (##list 'define (global-var->identifier (^ 1))
                    (##decomp (^ 0)))))

(define ##degen-if2
  (mk-degen ()
    (##list 'if (##decomp (^ 0))
                (##decomp (^ 1)))))

(define ##degen-if3
  (mk-degen ()
    (##list 'if (##decomp (^ 0))
                (##decomp (^ 1))
                (##decomp (^ 2)))))

(define ##degen-seq
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (if (##begin? val2)
        (##cons 'begin (##cons val1 (##cdr val2)))
        (##list 'begin val1 val2)))))

(define ##degen-quasi-list->vector
  (mk-degen ()
    (##list 'quasiquote (##vector (##list 'unquote-splicing (##decomp (^ 0)))))))

(define ##degen-quasi-append
  (mk-degen ()
    (##list 'quasiquote (##list (##list 'unquote-splicing (##decomp (^ 0)))
                                (##list 'unquote-splicing (##decomp (^ 1)))))))

(define ##degen-quasi-cons
  (mk-degen ()
    (##list 'quasiquote (##list (##list 'unquote (##decomp (^ 0)))
                                (##list 'unquote-splicing (##decomp (^ 1)))))))

(define ##degen-cond-if
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1)))
          (val3 (##decomp (^ 2))))
      (##build-cond
        (if (##begin? val2) (##cons val1 (##cdr val2)) (##list val1 val2))
        val3))))

(define ##degen-cond-or
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (##build-cond (##list val1) val2))))

(define ##degen-cond-send
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1)))
          (val3 (##decomp (^ 2))))
      (##build-cond (##list val1 '=> val2) val3))))

(define (##build-cond clause rest)
  (cond ((##cond? rest)
         (##cons 'cond (##cons clause (##cdr rest))))
        ((##begin? rest)
         (##cons 'cond (##list clause (##cons 'else (##cdr rest)))))
        ((##void-constant? rest)
         (##list 'cond clause))
        (else
         (##list 'cond clause (##list 'else rest)))))

(define ##degen-or
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (if (##or? val2)
        (##cons 'or (##cons val1 (##cdr val2)))
        (##list 'or val1 val2)))))

(define ##degen-and
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (if (##and? val2)
        (##cons 'and (##cons val1 (##cdr val2)))
        (##list 'and val1 val2)))))

(define ##degen-case
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (##cons 'case (##cons val1 val2)))))

(define ##degen-case-clause
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (##cons (if (##begin? val1)
                (##cons (^ 2) (##cdr val1))
                (##list (^ 2) val1))
              val2))))

(define ##degen-case-else
  (mk-degen ()
    (let ((val (##decomp (^ 0))))
      (if (##void-constant? val)
        '()
        (##list (if (##begin? val)
                  (##cons 'else (##cdr val))
                  (##list 'else val)))))))

(define ##degen-let
  (mk-degen ()
    (let ((n (code-length $code)))
      (let loop ((i (##fixnum.- n 2)) (vals '()))
        (if (##fixnum.< 0 i)
          (loop (##fixnum.- i 1)
                (##cons (##decomp (code-ref $code i)) vals))
          (let ((body (##decomp (^ 0)))
                (bindings (##make-bindings (code-ref $code (##fixnum.- n 1)) vals)))
            (if (##begin? body)
              (##cons 'let (##cons bindings (##cdr body)))
              (##list 'let bindings body))))))))

(define (##make-bindings l1 l2)
  (if (##pair? l1)
    (##cons (##list (##car l1) (##car l2))
            (##make-bindings (##cdr l1) (##cdr l2)))
    '()))

(define ##degen-letrec
  (mk-degen ()
    (let ((n (code-length $code)))
      (let loop ((i (##fixnum.- n 2)) (vals '()))
        (if (##fixnum.< 0 i)
          (loop (##fixnum.- i 1)
                (##cons (##decomp (code-ref $code i)) vals))
          (let ((body (##decomp (^ 0)))
                (bindings (##make-bindings (code-ref $code (##fixnum.- n 1)) vals)))
            (if (##begin? body)
              (##cons 'letrec (##cons bindings (##cdr body)))
              (##list 'letrec bindings body))))))))

(define ##degen-prc-req
  (mk-degen ()
    (let* ((n (code-length $code))
           (body (##decomp (^ 0)))
           (params (code-ref $code (##fixnum.- n 1))))
      (if (##begin? body)
        (##cons 'lambda (##cons params (##cdr body)))
        (##list 'lambda params body)))))

(define ##degen-prc-rest
  (mk-degen ()
    (let ((body (##decomp (^ 0)))
          (params (##make-params (^ 2) #t #f '())))
      (if (##begin? body)
        (##cons 'lambda (##cons params (##cdr body)))
        (##list 'lambda params body)))))

(define ##degen-prc
  (mk-degen ()
    (let ((n (code-length $code)))
      (let loop ((i (##fixnum.- n 7)) (inits '()))
        (if (##not (##fixnum.< i 1))
          (loop (##fixnum.- i 1)
                (##cons (##decomp (code-ref $code i)) inits))
          (let ((body
                 (##decomp (^ 0)))
                (params
                 (##make-params
                   (code-ref $code (##fixnum.- n 1))
                   (code-ref $code (##fixnum.- n 3))
                   (code-ref $code (##fixnum.- n 2))
                   inits)))
            (if (##begin? body)
              (##cons 'lambda (##cons params (##cdr body)))
              (##list 'lambda params body))))))))

(define (##make-params parms rest? keys inits)
  (let* ((nb-parms
          (##length parms))
         (nb-inits
          (##length inits))
         (nb-reqs
          (##fixnum.- nb-parms (##fixnum.+ nb-inits (if rest? 1 0))))
         (nb-opts
          (##fixnum.- nb-inits (if keys (##vector-length keys) 0))))

    (define (build-reqs)
      (let loop ((parms parms)
                 (i nb-reqs))
        (if (##fixnum.= i 0)
          (build-opts parms)
          (let ((parm (##car parms)))
            (##cons parm
                    (loop (##cdr parms)
                          (##fixnum.- i 1)))))))

    (define (build-opts parms)
      (if (##fixnum.= nb-opts 0)
        (build-rest parms inits)
        (##cons #!optional
                (let loop ((parms parms)
                           (i nb-opts)
                           (inits inits))
                  (if (##fixnum.= i 0)
                    (build-rest parms inits)
                    (let ((parm (##car parms))
                          (init (##car inits)))
                      (##cons (if (##eq? init #f) parm (##list parm init))
                              (loop (##cdr parms)
                                    (##fixnum.- i 1)
                                    (##cdr inits)))))))))

    (define (build-rest parms inits)
      (if rest?
        (let ((parm (##car parms)))
          (if (and (##fixnum.= nb-opts 0) (##not keys))
            parm
            (##cons #!rest (##cons parm (build-keys (##cdr parms) inits)))))
        (build-keys parms inits)))

    (define (build-keys parms inits)
      (if (##not keys)
        '()
        (##cons #!key
                (let loop ((parms parms)
                           (i (##vector-length keys))
                           (inits inits))
                  (if (##fixnum.= i 0)
                    '()
                    (let ((parm (##car parms))
                          (init (##car inits)))
                      (##cons (if (##eq? init #f) parm (##list parm init))
                              (loop (##cdr parms)
                                    (##fixnum.- i 1)
                                    (##cdr inits)))))))))

    (build-reqs)))

(define ##degen-app0
  (mk-degen ()
    (##list (##decomp (^ 0)))))

(define ##degen-app1
  (mk-degen ()
    (##list (##decomp (^ 0))
            (##decomp (^ 1)))))

(define ##degen-app2
  (mk-degen ()
    (##list (##decomp (^ 0))
            (##decomp (^ 1))
            (##decomp (^ 2)))))

(define ##degen-app3
  (mk-degen ()
    (##list (##decomp (^ 0))
            (##decomp (^ 1))
            (##decomp (^ 2))
            (##decomp (^ 3)))))

(define ##degen-app
  (mk-degen ()
    (let ((n (code-length $code)))
      (let loop ((i (##fixnum.- n 1)) (vals '()))
        (if (##not (##fixnum.< i 0))
          (loop (##fixnum.- i 1)
                (##cons (##decomp (code-ref $code i)) vals))
          vals)))))

(define ##degen-delay
  (mk-degen ()
    (##list 'delay (##decomp (^ 0)))))

(define ##degen-future
  (mk-degen ()
    (##list 'future (##decomp (^ 0)))))

;------------------------------------------------------------------------------

(define ##decomp-dispatch-table #f)

(define (##setup-decomp-dispatch-table)
  (set! ##decomp-dispatch-table
    (##list
      (##cons ##cprc-top         ##degen-top)

      (##cons ##cprc-cst-null    (mk-degen () (degen ##degen-cst-x '())))
      (##cons ##cprc-cst-true    (mk-degen () (degen ##degen-cst-x #t)))
      (##cons ##cprc-cst-false   (mk-degen () (degen ##degen-cst-x #f)))
      (##cons ##cprc-cst--2      (mk-degen () (degen ##degen-cst-x -2)))
      (##cons ##cprc-cst--1      (mk-degen () (degen ##degen-cst-x -1)))
      (##cons ##cprc-cst-0       (mk-degen () (degen ##degen-cst-x 0)))
      (##cons ##cprc-cst-1       (mk-degen () (degen ##degen-cst-x 1)))
      (##cons ##cprc-cst-2       (mk-degen () (degen ##degen-cst-x 2)))
      (##cons ##cprc-cst         ##degen-cst)

      (##cons ##cprc-loc-ref-0-1 (mk-degen () (degen ##degen-loc-ref-x-y 0 1)))
      (##cons ##cprc-loc-ref-0-2 (mk-degen () (degen ##degen-loc-ref-x-y 0 2)))
      (##cons ##cprc-loc-ref-0-3 (mk-degen () (degen ##degen-loc-ref-x-y 0 3)))
      (##cons ##cprc-loc-ref-1-1 (mk-degen () (degen ##degen-loc-ref-x-y 1 1)))
      (##cons ##cprc-loc-ref-1-2 (mk-degen () (degen ##degen-loc-ref-x-y 1 2)))
      (##cons ##cprc-loc-ref-1-3 (mk-degen () (degen ##degen-loc-ref-x-y 1 3)))
      (##cons ##cprc-loc-ref-2-1 (mk-degen () (degen ##degen-loc-ref-x-y 2 1)))
      (##cons ##cprc-loc-ref-2-2 (mk-degen () (degen ##degen-loc-ref-x-y 2 2)))
      (##cons ##cprc-loc-ref-2-3 (mk-degen () (degen ##degen-loc-ref-x-y 2 3)))
      (##cons ##cprc-loc-ref     ##degen-loc-ref)
      (##cons ##cprc-glo-ref     ##degen-glo-ref)

      (##cons ##cprc-loc-set     ##degen-loc-set)
      (##cons ##cprc-glo-set     ##degen-glo-set)
      (##cons ##cprc-glo-def     ##degen-glo-def)

      (##cons ##cprc-if2         ##degen-if2)
      (##cons ##cprc-if3         ##degen-if3)
      (##cons ##cprc-seq         ##degen-seq)
      (##cons ##cprc-quasi-list->vector ##degen-quasi-list->vector)
      (##cons ##cprc-quasi-append ##degen-quasi-append)
      (##cons ##cprc-quasi-cons  ##degen-quasi-cons)
      (##cons ##cprc-cond-if     ##degen-cond-if)
      (##cons ##cprc-cond-or     ##degen-cond-or)
      (##cons ##cprc-cond-send-red ##degen-cond-send)
      (##cons ##cprc-cond-send-sub ##degen-cond-send)

      (##cons ##cprc-or          ##degen-or)
      (##cons ##cprc-and         ##degen-and)

      (##cons ##cprc-case        ##degen-case)
      (##cons ##cprc-case-clause ##degen-case-clause)
      (##cons ##cprc-case-else   ##degen-case-else)

      (##cons ##cprc-let         ##degen-let)
      (##cons ##cprc-letrec      ##degen-letrec)

      (##cons ##cprc-prc-req0    ##degen-prc-req)
      (##cons ##cprc-prc-req1    ##degen-prc-req)
      (##cons ##cprc-prc-req2    ##degen-prc-req)
      (##cons ##cprc-prc-req3    ##degen-prc-req)
      (##cons ##cprc-prc-req     ##degen-prc-req)
      (##cons ##cprc-prc-rest    ##degen-prc-rest)
      (##cons ##cprc-prc         ##degen-prc)

      (##cons ##cprc-app0-red    ##degen-app0)
      (##cons ##cprc-app1-red    ##degen-app1)
      (##cons ##cprc-app2-red    ##degen-app2)
      (##cons ##cprc-app3-red    ##degen-app3)
      (##cons ##cprc-app-red     ##degen-app)
      (##cons ##cprc-app0-sub    ##degen-app0)
      (##cons ##cprc-app1-sub    ##degen-app1)
      (##cons ##cprc-app2-sub    ##degen-app2)
      (##cons ##cprc-app3-sub    ##degen-app3)
      (##cons ##cprc-app-sub     ##degen-app)

      (##cons ##cprc-delay       ##degen-delay)
      (##cons ##cprc-future      ##degen-future)
)))

(##setup-decomp-dispatch-table)

;------------------------------------------------------------------------------

(define (##decomp $code)
  (let ((cprc (code-cprc $code)))
    (let ((x (##assq cprc ##decomp-dispatch-table)))
      (if x
        (degen (##cdr x))
        '?))))

(define (##decompile proc)

  (define (decomp proc)
    (let ((parent (##subprocedure-parent proc))
          (id (##subprocedure-id proc)))
      (let ((info (##procedure-info parent)))
        (if info
          (let ((v (##vector-ref info 0)))
            (let loop ((i (##fixnum.- (##vector-length v) 1)))
              (if (##fixnum.< i 0)
                proc
                (let ((x (##vector-ref v i)))
                  (if (##fixnum.= id (##vector-ref x 0))
                    (source->expression (##vector-ref x 1))
                    (loop (##fixnum.- i 1)))))))
          proc))))

  (define (source-code x)
    (##vector-ref x 0))

  (define (source->expression source)

    (define (list->expression l)
      (cond ((##pair? l)
             (##cons (source->expression (##car l))
                     (list->expression (##cdr l))))
            ((##null? l)
             '())
            (else
             (source->expression l))))

    (define (vector->expression v)
      (let* ((len (##vector-length v))
             (x (##make-vector len #f)))
        (let loop ((i (##fixnum.- len 1)))
          (if (##not (##fixnum.< i 0))
            (begin
              (##vector-set! x i (source->expression (##vector-ref v i)))
              (loop (##fixnum.- i 1)))))
        x))

    (let ((code (source-code source)))
      (cond ((##pair? code)   (list->expression code))
            ((##vector? code) (vector->expression code))
            (else             code))))

  (cond ((##interp-procedure? proc)
         (##decomp (##interp-procedure-code proc)))
        ((##closure? proc)
         (decomp (##closure-code proc)))
        (else
         (decomp proc))))

;==============================================================================

; Utilities

;------------------------------------------------------------------------------

; Internal variables are uninteresting for the user.

(define (##internal-var? var)
  (or (##eq? var (self-var))
      (##eq? var (selector-var))
      (##eq? var (do-loop-var))))

;------------------------------------------------------------------------------

; Access to structure of closures for interpreter procedures.
;
; Layout of closed variables for closures created by ##cprc-prcXXX:
;
;   slot 1: $code
;   slot 2: proc
;   slot 3: rte

(define ##interp-procedure-code-pointers
  (##list (##closure-code (##cprc-prc-req0 #f #f))
          (##closure-code (##cprc-prc-req1 #f #f))
          (##closure-code (##cprc-prc-req2 #f #f))
          (##closure-code (##cprc-prc-req3 #f #f))
          (##closure-code (##cprc-prc-req  #f #f))
          (##closure-code (##cprc-prc-rest #f #f))
          (##closure-code (##cprc-prc      #f #f))))

(define (##interp-procedure? x)
  (and (##procedure? x)
       (##closure? x)
       (##memq (##closure-code x) ##interp-procedure-code-pointers)))

(define (##interp-procedure-code x) ; return "$code" closed variable of x
  (##closure-ref x 1))

(define (##interp-procedure-rte x) ; return "rte" closed variable of x
  (##closure-ref x 3))

;------------------------------------------------------------------------------

; Access to continuation frames

(define (##frame-creator f)
  (##subprocedure-parent (##frame-ret f)))

(define (##interp-frame? f)
  (let ((creator (##frame-creator f)))
    (or (##eq? creator ##subproblem-apply0)
        (##eq? creator ##subproblem-apply1)
        (##eq? creator ##subproblem-apply2)
        (##eq? creator ##subproblem-apply3)
        (##eq? creator ##subproblem-apply))))

(define (##interp-frame-code f)
  (let ((creator (##frame-creator f)))
    (if (##eq? creator ##subproblem-apply0)
      (##frame-stk-ref f 2)
      (##frame-stk-ref f 1))))

(define (##interp-frame-rte f)
  (let ((creator (##frame-creator f)))
    (if (or (##eq? creator ##subproblem-apply2)
            (##eq? creator ##subproblem-apply3))
      (##frame-stk-ref f 2)
      (##frame-stk-ref f 3))))

(define (##interp-frame-creator f)
  (let ((code (##interp-frame-code f))
        (rte (##interp-frame-rte f)))
    (##extract-container code rte)))

;------------------------------------------------------------------------------

; Access to run time environments

(define (##rte-shape code)

  (define (get-shape cte)
    (cond ((##cte-top? cte)
           #f)
          ((##cte-frame? cte)
           (##list->vector
             (##cons (get-shape (##cte-parent-cte cte))
                     (##cte-frame-vars cte))))
          (else
           (get-shape (##cte-parent-cte cte)))))

  (get-shape (code-cte code)))

(define (##rte-var-ref rte up over)
  (let loop ((r rte) (i up))
    (if (##fixnum.< 0 i)
      (loop (rte-up r) (##fixnum.- i 1))
      (##vector-ref r over))))

(define (##rte-var-set! rte up over val)
  (let loop ((r rte) (i up))
    (if (##fixnum.< 0 i)
      (loop (rte-up r) (##fixnum.- i 1))
      (##vector-set! r over val))))

;------------------------------------------------------------------------------

(define (##interesting-frame? f)
  (let ((creator (##frame-creator f)))
    (##not (or (##assq creator ##decomp-dispatch-table)
               (##eq? creator ##dynamic-env-bind)
               (##eq? creator ##read-eval-print)))))

(define (##continuation->subproblems cont)
  (let loop ((f (##continuation->frame cont)) (l '()))
    (if f
      (if (##interesting-frame? f)
        (loop (##frame-next f) (##cons f l))
        (loop (##frame-next f) l))
      (##reverse l))))

(define (##eval-in-frame expr f dyn-bindings)
  (let ((dyn-env (##cons dyn-bindings (##frame-dyn-env f))))
    (if (##interp-frame? f)
      (let* ((code (##interp-frame-code f))
             (cte (code-cte code))
             (rte (##interp-frame-rte f)))
        (##eval-inner expr cte rte dyn-env))
      (##eval-top-with-dyn-env expr ##interaction-cte dyn-env))))

(define (##procedure-name p)
  (or (##object->global-var->identifier p) p))

;------------------------------------------------------------------------------

; Read eval print loop

(define ##repl-write #f)
(set! ##repl-write #f)

(define ##repl-read #f)
(set! ##repl-read #f)

(define (##repl
          #!optional (in ##stdin) (out ##stdout) (prompt2 "> ") (prompt1 ""))
  (##call-with-current-continuation
    (lambda (cont) (##read-eval-print in out prompt2 prompt1 cont))))

(define (##read-eval-print in out prompt2 prompt1 cont)

  (define (repl-start subprobs repl-info dyn-bindings)

    (define (repl-read)
      (let ((proc ##repl-read))
        (if (##procedure? proc)
          (proc in)
          (##read in))))

    (define (repl-write val)
      (let ((proc ##repl-write))
        (if (##procedure? proc)
          (proc val out)
          (if (##not (##eq? val (##void)))
            (begin
              (##write val out (if-forces #t #f))
              (##newline out))))))

    (define (repl-n n)
      (let loop ((i 0) (s subprobs))
        (if (and (##fixnum.< n i) (##pair? (##cdr s)))
          (loop (##fixnum.- i 1) (##cdr s))
          (let ((f (##car s)))
            (##display-subproblem i f out)
            (repl i s f)))))

    (define (cmd-d eof?)
      (let ((l (##cdr (##vector-ref repl-info 3))))
        (if eof?
          (##newline out))
        (cond ((##pair? l)
               ((##car l) #f))
              (eof?
               (##write-string "*** EOF again to exit" out)
               (##newline out)
               (if (##eof-object? (##peek-char in))
                 (##exit))))))

    (define (cmd-t)
      (let loop ((l (##vector-ref repl-info 3)))
        (if (##pair? (##cdr l))
          (loop (##cdr l))
          ((##car l) #f))))

    (define (repl pos subprobs* f)

      (##call-with-current-continuation
        (lambda (abort)
          (##set-car! (##vector-ref repl-info 3) abort)))

      (let loop ()

        (##display prompt1 out #f)
        (if (##fixnum.< pos 0) (##display pos out #f))
        (##display prompt2 out #f)

        (let ((expr (repl-read)))
          (if (##eof-object? expr)
            (begin (cmd-d #t) (loop))
            (if (and (##pair? expr)
                     (##pair? (##cdr expr))
                     (##null? (##cddr expr))
                     (##eq? (##car expr) 'unquote))
              (let ((cmd (##cadr expr)))
                (if (##eof-object? cmd)
                  (begin (cmd-d #t) (loop))
                  (case cmd
                    ((?) (##cmd-? out) (loop))
                    ((-) (repl-n (##fixnum.- pos 1)))
                    ((+) (repl-n (##fixnum.+ pos 1)))
                    ((b) (##cmd-b pos subprobs* out) (loop))
                    ((i) (##cmd-i f out) (loop))
                    ((y) (##cmd-y f out) (loop))
                    ((l) (##cmd-l f out) (loop))
                    ((t) (cmd-t))
                    ((d) (cmd-d #f) (loop))
                    ((r) (##display "Return value: " out #f)
                         (let ((expr (repl-read)))
                           (if (##eof-object? expr)
                             (##void)
                             (##eval-in-frame expr f dyn-bindings))))
                    ((q) (##exit))
                    (else
                     (if (and (##fixnum? cmd) (##fixnum.< cmd 1))
                       (repl-n cmd)
                       (begin
                         (##write-string "Unknown command ," out)
                         (##write cmd out #f)
                         (##newline out)
                         (loop)))))))
            (let ((val (##eval-in-frame expr f dyn-bindings)))
              (repl-write val)
              (loop)))))))

    (repl 0 subprobs (##car subprobs)))

  (let* ((prev-info (##dynamic-ref '##repl-info #f))
         (repl-info
           (##vector
             in
             out
             (if prev-info
               (##fixnum.+ (##vector-ref prev-info 2) 1)
               0)
             (if prev-info
               (let ((conts (##vector-ref prev-info 3)))
                 (##cons (##car conts) conts))
               (##cons (lambda (x) (##exit 1)) '()))))
         (dyn-bindings (##list (##cons '##repl-info repl-info))))
    (##dynamic-let
      dyn-bindings
      (lambda ()
        (repl-start (##continuation->subproblems cont)
                    repl-info
                    dyn-bindings)))))

(define (##repl-out)
  (let ((repl-info (##dynamic-ref '##repl-info #f)))
    (if repl-info
      (##vector-ref repl-info 1)
      ##stderr)))

(define (##debug-repl cont)
  (let ((repl-info (##dynamic-ref '##repl-info #f)))
    (if repl-info
      (##read-eval-print (##vector-ref repl-info 0)
                         (##vector-ref repl-info 1)
                         "> "
                         (##fixnum.+ (##vector-ref repl-info 2) 1)
                         cont)
      (##exit 1))))

(define (##pop-repl)
  (let ((repl-info (##dynamic-ref '##repl-info #f)))
    (if repl-info
      ((##car (##vector-ref repl-info 3)) #f)
      (##exit 1))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##cmd-? out)
  (##write-string ",?        : Summary of commands" out) (##newline out)
  (##write-string ",+ and ,- : Move to next or previous frame of continuation" out) (##newline out)
  (##write-string ",<n>      : Move to particular frame (<n> <= 0)" out) (##newline out)
  (##write-string ",b        : Display frames of continuation (i.e. backtrace)" out) (##newline out)
  (##write-string ",i        : Display procedure attached to current frame" out) (##newline out)
  (##write-string ",y        : Display subproblem of current frame" out) (##newline out)
  (##write-string ",l        : Display list of local variables accessible in current frame" out) (##newline out)
  (##write-string ",t        : Transfer to top-level REP loop" out) (##newline out)
  (##write-string ",d        : Transfer to previous REP loop" out) (##newline out)
  (##write-string ",r        : Return from REP loop" out) (##newline out)
  (##write-string ",q        : Quit" out) (##newline out))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##cmd-b pos subprobs* out)
  (define max-head 10)
  (define max-tail 6)
  (let loop ((i 0) (j (##fixnum.- (##length subprobs*) 1)) (l subprobs*))
    (if (##pair? l)
      (begin
        (cond ((or (##fixnum.< i max-head) (##fixnum.< j max-tail)
                   (and (##fixnum.= i max-head) (##fixnum.= j max-tail)))
               (##display-subproblem (##fixnum.- pos i) (##car l) out))
              ((##fixnum.= i max-head)
               (##write-string "..." out) (##newline out)))
        (loop (##fixnum.+ i 1) (##fixnum.- j 1) (##cdr l))))))

(define (##display-subproblem pos f out)

  (define (display-subproblem-call x call)
    (if call
      (begin
        (##display-spaces (##fixnum.- 25 x) out)
        (##write-string " " out)
        (##write-string (##object->string call 48 #f) out))))

  (let ((x (##write pos out #f)))
    (##display-spaces (##fixnum.- 4 x) out)
    (##write-string " " out)
    (if (##interp-frame? f)

      (let ((creator (##interp-frame-creator f)))
        (display-subproblem-call
          (if creator
            (##write (##procedure-name creator) out #f)
            (##display "(interaction)" out #f))
          (##decomp (##interp-frame-code f))))

      (let* ((creator (##frame-creator f))
             (ret (##frame-ret f))
             (call (##decompile ret)))
        (display-subproblem-call
          (##write (##procedure-name creator) out #f)
          (if (##eq? call ret) #f call))))

    (##newline out)))

(define (##display-spaces n out)
  (if (##fixnum.< 0 n)
    (let ((m (if (##fixnum.< 40 n) 40 n)))
      (##write-substring "                                        " 0 m out)
      (##display-spaces (##fixnum.- n m) out))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##cmd-l f out)

  (define (display-locals cte rte)
    (let loop1 ((c cte) (r rte))
      (cond ((##cte-top? c))
            ((##cte-frame? c)
             (let loop2 ((vars (##cte-frame-vars c))
                         (vals (##cdr (##vector->list r))))
               (if (##pair? vars)
                 (let ((var (##car vars)))
                   (if (##not (##internal-var? var))
                     (let ((x (##write var out #f)))
                       (##write-string " = " out)
                       (##write-string
                        (##object->string
                          (##car vals)
                          (##fixnum.- (##fixnum.- (##port-width out) 3) x)
                          (if-forces #t #f))
                        out)
                       (##newline out)))
                   (loop2 (##cdr vars) (##cdr vals)))))
             (loop1 (##cte-parent-cte c) (rte-up r)))
            (else
             (loop1 (##cte-parent-cte c) r)))))

  (if (##interp-frame? f)
    (display-locals (code-cte (##interp-frame-code f))
                    (##interp-frame-rte f))
    (begin
      (##write-string "Sorry, can't display compiled code environment" out)
      (##newline out))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##cmd-y f out)
  (if (##interp-frame? f)
    (##pretty-print (##decomp (##interp-frame-code f)) out (##port-width out))
    (let ((x (##decompile (##frame-ret f))))
      (if (##eq? x (##frame-ret f))
        (begin
          (##write-string "Sorry, this code was compiled without the DEBUG option" out)
          (##newline out))
        (##pretty-print x out (##port-width out))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##cmd-i f out)
  (if (##interp-frame? f)

    (let ((creator (##interp-frame-creator f)))
      (if creator
        (begin
          (##write creator out #f)
          (##write-string " =" out)
          (##newline out)
          (##pretty-print (##decompile creator) out (##port-width out)))
        (begin
          (##write-string "(interaction)" out)
          (##newline out))))

    (let ((creator (##frame-creator f)))
      (##write creator out #f)
      (let ((x (##decompile creator)))
        (if (##eq? x creator)
          (##newline out)
          (begin
            (##write-string " =" out)
            (##newline out)
            (##pretty-print x out (##port-width out))))))))

;==============================================================================
