/* file: "main.c" */

/* Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved. */

/* This is the driver of the Gambit-C system */

#define ___INCLUDED_FROM_MAIN
#define ___VERSION 19
#include "gambit.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>


___HIDDEN void usage_err (argv)
char *argv[];
{
  fprintf (stderr, "Usage: %s [-:OPT,OPT...] ...\n", argv[0]);
  fprintf (stderr, "  where OPT can be one of:\n");
  fprintf (stderr, "    mMIN_HEAP_SIZE_IN_KILOBYTES\n");
  fprintf (stderr, "    hMAX_HEAP_SIZE_IN_KILOBYTES\n");
  fprintf (stderr, "    s\n");
  fprintf (stderr, "    d\n");
  fprintf (stderr, "    t\n");
  fprintf (stderr, "    c\n");
  fprintf (stderr, "    1\n");
  fprintf (stderr, "    8\n");
  exit (1);
}


___EXP(int,___main) (argc, argv, linker)
int argc;
char **argv;
___mod_or_lnk (*linker)();
{
  unsigned long min_heap_len_in_k, max_heap_len_in_k;
  int default_io_encoding;
  int standard;
  int debug_level;
  int force_tty;

  /* handle arguments to runtime */

  min_heap_len_in_k = 0;
  max_heap_len_in_k = 0;
  default_io_encoding = ___DEFAULT_IO_ENCODING;
  standard = 0;
  debug_level = 0;
  force_tty = 0;

  if (argc > 1)
    {
      char *s, *arg = argv[1];
      if (*arg++ == '-' && *arg++ == ':')
        {
          do
            {
              s = arg++;
              switch (*s)
                {
                case 'm':
                case 'h':
                  {
                    char temp;
                    int argval;
                    while (isdigit (*arg)) arg++;
                    temp = *arg;
                    *arg = '\0';
                    argval = atoi (s+1);
                    *arg = temp;
                    if (*s == 'h')
                      max_heap_len_in_k = argval;
                    else
                      min_heap_len_in_k = argval;
                    break;
                  }
                case 's':
                  standard = 1;
                  break;
                case 'd':
                  debug_level = 1;
                  break;
                case 't':
                  force_tty = 1;
                  break;
                case 'c':
                  default_io_encoding = ___IO_CHAR_ENCODING;
                  break;
                case '1':
                  default_io_encoding = ___IO_LATIN1_ENCODING;
                  break;
                case '8':
                  default_io_encoding = ___IO_UTF8_ENCODING;
                  break;
                default:
                  usage_err (argv);
                }
            }
          while (*arg++ == ',');
          if (*--arg != '\0')
            usage_err (argv);
          argv[1] = argv[0];
          argc--;
          argv++;
        }
    }

  if (min_heap_len_in_k > (unsigned long)-1>>10)
    min_heap_len_in_k = (unsigned long)-1>>10;

  if (max_heap_len_in_k > (unsigned long)-1>>10)
    max_heap_len_in_k = (unsigned long)-1>>10;

  /* Setup program, run it and perform any cleanup necessary. */

  {
    ___setup_params setup_params;

    setup_params.argc                = argc;
    setup_params.argv                = argv;
    setup_params.min_heap            = min_heap_len_in_k << 10;
    setup_params.max_heap            = max_heap_len_in_k << 10;
    setup_params.gc_hook             = 0;
    setup_params.fatal_error         = 0;
    setup_params.standard            = standard;
    setup_params.debug_level         = debug_level;
    setup_params.default_io_encoding = default_io_encoding;
    setup_params.force_tty           = force_tty;
    setup_params.linker              = linker;

    ___setup (&setup_params);

    ___cleanup ();
  }

  /* Exit program. */

  exit (0);

  return 0; /* satisfy some compilers */
}
