/* file: "mem.h" */

/* Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved. */

/*
 * Memory allocation parameters:
 *
 * ___MAX_NB_PARMS is the maximum number of formal parameters of procedures.
 *
 * ___MAX_NB_ARGS is the maximum number of arguments (actual parameters)
 * that are passed to procedures.  It must be >= ___MAX_NB_PARMS.
 *
 * ___MAX_NB_FRAME_SLOTS is the maximum number of slots in continuation
 * frames.  It must be >= ___MAX_NB_ARGS.
 *
 * ___STACK_NON_FUDGE is the size in words of the stack, not counting
 * the stack fudge area.  It must be >= ___MAX_NB_FRAME_SLOTS.
 *
 * ___STACK_FUDGE is the size in words of the stack fudge area which is a
 * special area at the end of the stack-cache.  Polling instructions in
 * the code generated by the compiler check every so often whether the frame
 * pointer points into the stack fudge area, which indicates a "stack-cache
 * overflow".  In this case the continuation frames in the stack-cache are
 * moved to the heap and the stack-cache is cleared, leaving room for
 * subsequent allocations on the stack.  There is at most one continuation
 * frame in the stack-cache.  Because of the way polling instructions are
 * added by the compiler and the way keyword parameters are handled,
 * ___STACK_FUDGE must be >= ___MAX_NB_ARGS+___MAX_NB_PARMS.
 *
 * ___NORMAL_OVERFLOW_RESERVE is the number of words reserved in the heap
 * for handling heap overflows.  When the heap overflows, this space will
 * be made available to the heap overflow handler.
 *
 * ___MSECTION_SIZE is the size in words of sections that contain
 * movable objects (msections).
 *
 * ___MSECTION_BIGGEST is the size in words of the biggest movable object.
 * It must be <= ___MSECTION_SIZE.
 *
 * ___MSECTION_SMALLEST is the size in words of smallest movable object.
 * It must be <= ___MSECTION_BIGGEST.
 *
 * ___MIN_NB_MSECTIONS is the minimum number of msections contained
 * in the heap.
 *
 * ___PSECTION_SIZE is the size in words of sections that contain permanent
 * objects (psections).
 *
 * ___PSECTION_BIGGEST is the size in words of biggest object in psections.
 * It must be <= ___PSECTION_SIZE.
 */

#define ___MAX_NB_PARMS       1024
#define ___MAX_NB_ARGS        4096
#define ___MAX_NB_FRAME_SLOTS 4096
#define ___STACK_NON_FUDGE    4096
#define ___STACK_FUDGE        (___MAX_NB_ARGS+___MAX_NB_PARMS)
#define ___NORMAL_OVERFLOW_RESERVE \
(___STACK_NON_FUDGE*(1+___FRAME_OVERHEAD) + (___STACK_FUDGE+___FRAME_OVERHEAD))
#define ___MSECTION_SIZE      131072
#define ___MSECTION_BIGGEST   4096
#define ___MSECTION_SMALLEST  0
#define ___MIN_NB_MSECTIONS   2
#define ___PSECTION_SIZE      1024
#define ___PSECTION_BIGGEST   32

extern void ___setup_mem (/* setup_params */);
extern void ___cleanup_mem ();

extern int ___gc ();
extern int ___heap_limit ();

extern ___glo_struct *___alloc_global_var ();
